#!/usr/bin/env python3
# License: BSD-3-Clause

import os, time, gzip, io
from datetime import datetime
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, MultipleLocator
from matplotlib.colors import Normalize
from base64 import b64decode

from astropy.time import Time
from astropy.cosmology import FlatLambdaCDM
from astropy import units as u
from astropy.io import fits
from astropy import visualization
from ztfquery.utils.stamps import get_ps_stamp
from nuztf.cat_match import get_cross_match_info


# For absolute magnitude calculation
GENERIC_COSMOLOGY = FlatLambdaCDM(H0=70, Om0=0.3)


def lightcurve_from_alert(
    alert: list,
    # figsize: list=[6.47, 4],
    figsize: list = [8, 5],
    title: str = None,
    include_ulims: bool = True,
    include_cutouts: bool = True,
    include_crossmatch: bool = True,
    mag_range: list = None,
    z: float = None,
    legend: bool = False,
    logger=None,
):
    """plot AMPEL alerts as lightcurve"""

    if logger is None:
        import logging

        logger = logging.getLogger(__name__)
    else:
        logger = logger

    if z is not None:
        if np.isnan(z):
            z = None
            logger.debug("Redshift is nan, will be ignored")

    # ZTF color and naming scheme
    BAND_NAMES = {1: "ZTF g", 2: "ZTF r", 3: "ZTF i"}
    BAND_COLORS = {1: "green", 2: "red", 3: "orange"}

    name = alert[0]["objectId"]
    candidate = alert[0]["candidate"]
    prv_candid = alert[0]["prv_candidates"]

    if include_cutouts:
        try:
            cutouts = alert[0]["cutouts"]
        except:
            logger.warning(
                "The alert dictionary does not contain cutouts. Will proceed without them."
            )
            include_cutouts = False

    logger.debug(f"Plotting {name}")
    logger.debug(f"Found {len(prv_candid)+1} alerts")

    df = pd.DataFrame(candidate, index=[0])
    df_ulims = pd.DataFrame()

    # Filter out images with negative difference flux
    i = 0
    for prv in prv_candid:
        # Go through the alert history
        if "magpsf" in prv.keys() and "isdiffpos" in prv.keys():
            i += 1
            ser = pd.Series(prv, name=i)
            df = df.append(ser)
        else:
            df_ulims = df_ulims.append(prv, ignore_index=True)
            i += 1

    df["mjd"] = df["jd"] - 2400000.5
    df_ulims["mjd"] = df_ulims["jd"] - 2400000.5

    # Helper functions for the axis conversion (from MJD to days from today)
    def t0_dist(obsmjd):
        t0 = Time(time.time(), format="unix", scale="utc").mjd
        return obsmjd - t0

    def t0_to_mjd(dist_to_t0):
        t0 = Time(time.time(), format="unix", scale="utc").mjd
        return t0 + dist_to_t0

    def mjd_to_date(mjd):
        # return mjd + 20000
        return Time(mjd, format="mjd").mjd

    def date_to_mjd(date):
        return Time(date, format="mjd").mjd

    fig = plt.figure(figsize=figsize)

    if include_cutouts:
        lc_ax1 = fig.add_subplot(5, 4, (9, 19))
        cutoutsci = fig.add_subplot(5, 4, (1, 5))
        cutouttemp = fig.add_subplot(5, 4, (2, 6))
        cutoutdiff = fig.add_subplot(5, 4, (3, 7))
        cutoutps1 = fig.add_subplot(5, 4, (4, 8))
    else:
        lc_ax1 = fig.add_subplot(1, 1, 1)
        fig.subplots_adjust(top=0.8, bottom=0.15)

    plt.subplots_adjust(wspace=0.4, hspace=1.8)

    #

    if include_cutouts:
        for cutout_, ax_, type_ in zip(
            [cutouts, cutouts, cutouts],
            [cutoutsci, cutouttemp, cutoutdiff],
            ["Science", "Template", "Difference"],
        ):
            create_stamp_plot(cutouts=cutout_, ax=ax_, type=type_)

        img = get_ps_stamp(
            candidate["ra"], candidate["dec"], size=240, color=["y", "g", "i"]
        )
        cutoutps1.imshow(np.asarray(img))
        cutoutps1.set_title("PS1", fontdict={"fontsize": "small"})
        cutoutps1.set_xticks([])
        cutoutps1.set_yticks([])

    # If redshift is given, calculate absolute magnitude via luminosity distance
    # and plot as right axis
    if z is not None:

        dist_l = GENERIC_COSMOLOGY.luminosity_distance(z).to(u.pc).value

        def mag_to_absmag(mag):
            absmag = mag - 5 * (np.log10(dist_l) - 1)
            return absmag

        def absmag_to_mag(absmag):
            mag = absmag + 5 * (np.log10(dist_l) - 1)
            return mag

        lc_ax3 = lc_ax1.secondary_yaxis(
            "right", functions=(mag_to_absmag, absmag_to_mag)
        )

        if not include_cutouts:
            lc_ax3.set_ylabel(f"Absolute Magnitude [AB]")

    # Give the figure a title
    if not include_cutouts:
        if title is None:
            fig.suptitle(f"{name}", fontweight="bold")
        else:
            fig.suptitle(title, fontweight="bold")

    # grid line every 100 days
    # lc_ax1.xaxis.set_major_locator(MultipleLocator(100))

    lc_ax1.grid(b=True, axis="both", alpha=0.5)
    lc_ax1.set_ylabel("Magnitude [AB]")

    if not include_cutouts:
        lc_ax1.set_xlabel("MJD")

    # Determine magnitude limits
    if mag_range is None:
        max_mag = np.max(df.magpsf.values) + 0.3
        min_mag = np.min(df.magpsf.values) - 0.3
        lc_ax1.set_ylim([max_mag, min_mag])
    else:
        lc_ax1.set_ylim([np.max(mag_range), np.min(mag_range)])

    for fid in BAND_NAMES.keys():

        # Plot older datapoints
        df_temp = df.iloc[1:].query("fid == @fid")
        lc_ax1.errorbar(
            df_temp["mjd"],
            df_temp["magpsf"],
            df_temp["sigmapsf"],
            color=BAND_COLORS[fid],
            fmt=".",
            label=BAND_NAMES[fid],
            mec="black",
            mew=0.5,
        )

        # Plot upper limits
        if include_ulims:
            df_temp2 = df_ulims.query("fid == @fid")
            lc_ax1.scatter(
                df_temp2["mjd"],
                df_temp2["diffmaglim"],
                c=BAND_COLORS[fid],
                marker="v",
                s=1.3,
                alpha=0.5,
            )

    # Plot datapoint from alert
    df_temp = df.iloc[0]
    fid = df_temp["fid"]
    lc_ax1.errorbar(
        df_temp["mjd"],
        df_temp["magpsf"],
        df_temp["sigmapsf"],
        color=BAND_COLORS[fid],
        fmt=".",
        label=BAND_NAMES[fid],
        mec="black",
        mew=0.5,
        markersize=12,
    )

    if legend:
        plt.legend()

    # Now we create an infobox
    if include_cutouts:
        info = []

        info.append(name)
        info.append("------------------------")
        info.append(f"RA: {candidate['ra']:.8f}")
        info.append(f"Dec: {candidate['dec']:.8f}")
        info.append(f"rb: {candidate['rb']:.3f}")
        info.append("------------------------")

        for kk in ["sgscore", "distpsnr", "srmag"]:
            for k in [k for k in candidate.keys() if kk in k]:
                info.append(f"{k}: {candidate.get(k):.3f}")

        fig.text(0.77, 0.55, "\n".join(info), va="top", fontsize="medium", color="0.4")

    if include_crossmatch:
        xmatch_info = get_cross_match_info(alert[0])
        fig.text(
            0.5,
            0.975,
            xmatch_info,
            va="top",
            ha="center",
            fontsize="medium",
            color="0.4",
        )

    # Ugly hack because secondary_axis does not work with astropy.time.Time datetime conversion

    mjd_min = np.min(df.mjd.values)
    mjd_max = np.max(df.mjd.values)
    length = mjd_max - mjd_min

    lc_ax1.set_xlim([mjd_min - (length / 20), mjd_max + (length / 20)])

    lc_ax2 = lc_ax1.twiny()

    datetimes = [Time(x, format="mjd").datetime for x in [mjd_min, mjd_max]]

    lc_ax2.scatter(
        [Time(x, format="mjd").datetime for x in [mjd_min, mjd_max]], [20, 20], alpha=0
    )
    lc_ax2.tick_params(axis="both", which="major", labelsize=6, rotation=45)
    lc_ax1.tick_params(axis="x", which="major", labelsize=9, rotation=45)
    lc_ax1.tick_params(axis="y", which="major", labelsize=9)

    if z is not None:
        lc_ax3.tick_params(axis="both", which="major", labelsize=9)

    if z is not None:
        axes = [lc_ax1, lc_ax2, lc_ax3]
    else:
        axes = [lc_ax1, lc_ax2]

    return fig, axes


def create_stamp_plot(cutouts: dict, ax, type: str):
    """Helper function to create cutout subplot"""
    with gzip.open(io.BytesIO(cutouts[f"cutout{type}"]["data"]), "rb") as f:
        data = fits.open(io.BytesIO(f.read()), ignore_missing_simple=True)[0].data
    vmin, vmax = np.percentile(data[data == data], [0, 100])
    data_ = visualization.AsinhStretch()((data - vmin) / (vmax - vmin))
    ax.imshow(
        data_,
        norm=Normalize(*np.percentile(data_[data_ == data_], [0.5, 99.5])),
        aspect="auto",
    )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(type, fontdict={"fontsize": "small"})



#!/usr/bin/env python3
# coding: utf-8

import os, time, json, logging, datetime

import backoff
import requests
import pandas
import healpy as hp
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

from astropy.time import Time
from astropy import units as u
from astropy.coordinates import SkyCoord, Distance
from astropy.cosmology import Planck18 as cosmo

from ztfquery import alert
from ztfquery import fields as ztfquery_fields

from gwemopt.ztf_tiling import get_quadrant_ipix

from ampel.ztf.t0.DecentFilter import DecentFilter
from ampel.ztf.dev.DevAlertProcessor import DevAlertProcessor
from ampel.alert.PhotoAlert import PhotoAlert

from nuztf.ampel_api import (
    ampel_api_cone,
    ampel_api_timerange,
    ampel_api_name,
    add_cutouts,
)
from nuztf.cat_match import get_cross_match_info, ampel_api_tns, query_ned_for_z
from nuztf.observation_log import get_obs_summary
from nuztf.plot import lightcurve_from_alert

DEBUG = False
RATELIMIT_CALLS = 10
RATELIMIT_PERIOD = 1


class BaseScanner:
    def __init__(
        self,
        run_config,
        t_min,
        resource=None,
        filter_class=DecentFilter,
        cone_nside=64,
        cones_to_scan=None,
        logger=None,
    ):
        self.cone_nside = cone_nside
        self.t_min = t_min

        if not hasattr(self, "prob_threshold"):
            self.prob_threshold = None

        if resource is None:
            resource = {
                "ampel-ztf/catalogmatch": "https://ampel.zeuthen.desy.de/api/catalogmatch/",
            }

        if logger is None:
            import logging

            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

        self.logger.info("AMPEL run config:")
        self.logger.info(run_config)

        self.ampel_filter_class = filter_class(
            logger=self.logger, resource=resource, **run_config
        )

        self.dap = DevAlertProcessor(self.ampel_filter_class)

        # if not hasattr(self, "summary_path"):
        #     self.summary_path = None

        self.scanned_pixels = []

        if cones_to_scan is None:
            self.cone_ids, self.cone_coords = self.find_cone_coords()
        else:
            self.cone_ids, self.cone_coords = cones_to_scan

        self.cache = dict()
        self.default_t_max = t_min + 10.0

        self.overlap_prob = None
        self.overlap_fields = None
        self.first_obs = None
        self.last_obs = None
        self.n_fields = None
        self.area = None
        self.double_extragalactic_area = None

        if not hasattr(self, "dist"):
            self.dist = None

    def get_name(self):
        raise NotImplementedError

    def get_full_name(self):
        raise NotImplementedError

    @staticmethod
    def get_tiling_line():
        return ""

    @staticmethod
    def get_obs_line():
        raise NotImplementedError

    @staticmethod
    def remove_variability_line():
        raise NotImplementedError

    def get_overlap_line(self):
        raise NotImplementedError

    def filter_ampel(self, res):
        return (
            self.ampel_filter_class.apply(
                PhotoAlert(res["objectId"], res["objectId"], *self.dap._shape(res))
            )
            is not None
        )

    # @sleep_and_retry
    # @limits(calls=RATELIMIT_CALLS, period=RATELIMIT_PERIOD)
    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )
    def get_avro_by_name(self, ztf_name):
        return ampel_api_name(ztf_name, logger=self.logger)

    def add_res_to_cache(self, res):

        for res_alert in res:

            if res_alert["objectId"] not in self.cache.keys():
                self.cache[res_alert["objectId"]] = res_alert
            elif (
                res_alert["candidate"]["jd"]
                > self.cache[res_alert["objectId"]]["candidate"]["jd"]
            ):
                self.cache[res_alert["objectId"]] = res_alert

    def add_to_cache_by_names(self, *args):
        for ztf_name in args:
            self.add_res_to_cache(self.get_avro_by_name(ztf_name))

    def check_ampel_filter(self, ztf_name):
        lvl = logging.getLogger().getEffectiveLevel()
        logging.getLogger().setLevel(logging.DEBUG)
        self.logger.info("Set logger level to DEBUG")
        all_query_res = self.get_avro_by_name(ztf_name)
        pipeline_bool = False
        for query_res in all_query_res:
            self.logger.info("Checking filter f (no prv)")
            no_prv_bool = self.filter_f_no_prv(query_res)
            self.logger.info(f"Filter f (np prv): {no_prv_bool}")
            if no_prv_bool:
                self.logger.info("Checking ampel filter")
                bool_ampel = self.filter_ampel(query_res)
                self.logger.info(f"ampel filter: {bool_ampel}")
                if bool_ampel:
                    self.logger.info("Checking filter f (history)")
                    history_bool = self.filter_f_history(query_res)
                    self.logger.info(f"Filter f (history): {history_bool}")
                    if history_bool:
                        pipeline_bool = True
        self.logger.info(f"Setting logger back to {lvl}")
        logging.getLogger().setLevel(lvl)
        return pipeline_bool

    def plot_ztf_observations(self, **kwargs):
        self.get_multi_night_summary().show_gri_fields(**kwargs)

    def get_multi_night_summary(self, max_days=None):
        return get_obs_summary(self.t_min, max_days=max_days)

    def scan_cones(self, t_max=None, max_cones: int = None):
        """ """
        if max_cones is None:
            max_cones = len(self.cone_ids)

        scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))

        self.logger.info("Commencing Ampel queries!")
        self.logger.info(f"Scan radius is {scan_radius}")
        self.logger.info(
            f"So far, {len(self.scanned_pixels)} pixels out of {len(self.cone_ids)} have already been scanned."
        )

        all_ztf_ids = []

        for i, cone_id in enumerate(tqdm(list(self.cone_ids)[:max_cones])):
            ra, dec = self.cone_coords[i]

            if cone_id not in self.scanned_pixels:
                ztf_ids = self.ampel_cone_search(
                    ra=np.degrees(ra),
                    dec=np.degrees(dec),
                    radius=scan_radius,
                    t_max=t_max,
                )

                if ztf_ids:
                    for ztf_id in ztf_ids:
                        all_ztf_ids.append(ztf_id)

                self.scanned_pixels.append(cone_id)

        self.logger.info(f"Scanned {len(self.scanned_pixels)} pixels")

        # remove duplicates
        all_ztf_ids = list(set(all_ztf_ids))

        self.logger.info(f"Before filtering: Found {len(all_ztf_ids)} candidates")
        self.logger.info(f"Retrieving alert history from AMPEL")

        results = self.ampel_object_search(ztf_ids=all_ztf_ids)

        for res in results:
            self.add_res_to_cache(res)

        self.logger.info(f"Found {len(self.cache)} candidates")

        self.create_candidate_summary()

    def filter_f_no_prv(self, res):
        raise NotImplementedError

    def fast_filter_f_no_prv(self, res):
        return self.filter_f_no_prv(res)

    def filter_f_history(self, res):
        raise NotImplementedError

    def fast_filter_f_history(self, res):
        return self.filter_f_history(res)

    def find_cone_coords(self):
        raise NotImplementedError

    @staticmethod
    def wrap_around_180(ra_deg: float):
        """ """
        ra_rad = np.deg2rad(ra_deg)
        ra_rad[ra_rad > np.pi] -= 2 * np.pi
        ra_deg = np.rad2deg(ra_rad)

        return ra_deg

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )
    def ampel_cone_search(
        self, ra: float, dec: float, radius: float, t_max=None
    ) -> list:
        """ """
        if t_max is None:
            t_max = self.default_t_max

        t_min = self.t_min

        query_res = ampel_api_cone(
            ra, dec, radius, t_min.jd, t_max.jd, logger=self.logger
        )

        ztf_names = []
        for res in query_res:
            if self.filter_f_no_prv(res):
                if self.filter_ampel(res):
                    ztf_names.append(res["objectId"])

        ztf_names = list(set(ztf_names))

        return ztf_names

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )
    def ampel_timerange_search(
        self,
        t_min=None,
        t_max=None,
        with_history: bool = False,
        chunk_size: int = 500,
    ) -> list:
        """ """
        if t_max is None:
            t_max = self.default_t_max

        self.t_min = t_min

        query_res = ampel_api_timerange(
            t_min_jd=t_min.jd,
            t_max_jd=t_max.jd,
            with_history=with_history,
            chunk_size=chunk_size,
            logger=self.logger,
        )

        ztf_names_unfiltered = []

        ztf_names = []

        for res in query_res:
            ztf_id = res["objectId"]
            ztf_names_unfiltered.append(ztf_id)
            if self.filter_f_no_prv(res):
                if self.filter_ampel(res):
                    ztf_names.append(ztf_id)

        ztf_names_unfiltered = list(set(ztf_names_unfiltered))
        ztf_names = list(set(ztf_names))

        return ztf_names_unfiltered

    def ampel_object_search(self, ztf_ids: list, with_history: bool = True) -> list:
        """ """
        all_results = []

        for ztf_id in ztf_ids:

            query_res = ampel_api_name(
                ztf_name=ztf_id, with_history=with_history, logger=self.logger
            )

            final_res = []

            for res in query_res:
                if self.filter_f_history(res):
                    final_res.append(res)

            all_results.append(final_res)

        return all_results

    @staticmethod
    def calculate_abs_mag(mag: float, redshift: float) -> float:
        """ """
        luminosity_distance = cosmo.luminosity_distance(redshift).value * 10 ** 6
        abs_mag = mag - 5 * (np.log10(luminosity_distance) - 1)

        return abs_mag

    def parse_candidates(self):

        table = (
            "+--------------------------------------------------------------------------------+\n"
            "| ZTF Name     | IAU Name  | RA (deg)    | DEC (deg)   | Filter | Mag   | MagErr |\n"
            "+--------------------------------------------------------------------------------+\n"
        )
        for name, res in sorted(self.cache.items()):

            jds = [x["jd"] for x in res["prv_candidates"]]

            if res["candidate"]["jd"] > max(jds):
                latest = res["candidate"]
            else:
                latest = res["prv_candidates"][jds.index(max(jds))]

            old_flag = ""

            second_det = [x for x in jds if x > min(jds) + 0.01]
            if len(second_det) > 0:
                if Time.now().jd - second_det[0] > 1.0:
                    old_flag = "(MORE THAN ONE DAY SINCE SECOND DETECTION)"

            tns_result = " ------- "
            tns_name, tns_date, tns_group = ampel_api_tns(
                latest["ra"], latest["dec"], searchradius_arcsec=3
            )
            if tns_name:
                tns_result = tns_name

            line = "| {0} | {1} | {2:011.7f} | {3:+011.7f} | {4}      | {5:.2f} | {6:.2f}   | {7} \n".format(
                name,
                tns_result,
                float(latest["ra"]),
                float(latest["dec"]),
                ["g", "r", "i"][latest["fid"] - 1],
                latest["magpsf"],
                latest["sigmapsf"],
                old_flag,
                sign="+",
                prec=7,
            )
            table += line

        table += "+--------------------------------------------------------------------------------+\n\n"
        return table

    def draft_gcn(self):

        first_obs_dt = self.first_obs.datetime
        pretty_date = first_obs_dt.strftime("%Y-%m-%d")
        pretty_time = first_obs_dt.strftime("%H:%M")

        text = (
            f"Astronomer Name (Institute of Somewhere), ............. report,\n"
            f"On behalf of the Zwicky Transient Facility (ZTF) and Global Relay of Observatories Watching Transients Happen (GROWTH) collaborations: \n"
            f"We observed the localization region of the {self.get_full_name()} with the Palomar 48-inch telescope, equipped with the 47 square degree ZTF camera (Bellm et al. 2019, Graham et al. 2019). {self.get_tiling_line()}"
            f"We started observations in the g-band and r-band beginning at {pretty_date} {pretty_time} UTC, "
            f"approximately {(self.first_obs.jd - self.t_min.jd) * 24.0:.1f} hours after event time. {self.get_overlap_line()}"
            f"{self.get_obs_line()} \n \n"
            "The images were processed in real-time through the ZTF reduction and image subtraction pipelines at IPAC to search for potential counterparts (Masci et al. 2019). "
            "AMPEL (Nordin et al. 2019, Stein et al. 2021) was used to search the alerts database for candidates. "
            "We reject stellar sources (Tachibana and Miller 2018) and moving objects, and "
            f"apply machine learning algorithms (Mahabal et al. 2019) {self.remove_variability_line()}. We are left with the following high-significance transient "
            "candidates by our pipeline, all lying within the "
            f"{100 * self.prob_threshold}% localization of the skymap.\n\n{self.parse_candidates()} \n\n"
        )

        if self.dist:
            text += (
                "The GW distance estimate is {:.0f} [{:.0f} - {:.0f}] Mpc.\n\n".format(
                    self.dist, self.dist - self.dist_unc, self.dist + self.dist_unc
                )
            )
        else:
            pass

        text += f"Amongst our candidates, \n{self.text_summary()}\n\n"

        text += (
            "ZTF and GROWTH are worldwide collaborations comprising Caltech, USA; IPAC, USA; WIS, Israel; OKC, Sweden; JSI/UMd, USA; DESY, Germany; TANGO, Taiwan; UW Milwaukee, USA; LANL, USA; TCD, Ireland; IN2P3, France.\n\n"
            "GROWTH acknowledges generous support of the NSF under PIRE Grant No 1545949.\n"
            "Alert distribution service provided by DIRAC@UW (Patterson et al. 2019).\n"
            "Alert database searches are done by AMPEL (Nordin et al. 2019).\n"
            "Alert filtering is performed with the AMPEL Follow-up Pipeline (Stein et al. 2021).\n"
        )
        if self.dist:
            text += "Alert filtering and follow-up coordination is being undertaken by the Fritz marshal system (FIXME CITATION NEEDED)."
        return text

    @staticmethod
    def extract_ra_dec(nside, index):
        """ """
        theta, phi = hp.pix2ang(nside, index, nest=True)
        ra_deg = np.rad2deg(phi)
        dec_deg = np.rad2deg(0.5 * np.pi - theta)

        return (ra_deg, dec_deg)

    @staticmethod
    def extract_npix(nside, ra, dec):
        """ " """
        theta = 0.5 * np.pi - np.deg2rad(dec)
        phi = np.deg2rad(ra)

        return hp.ang2pix(nside, theta, phi, nest=True)

    def create_candidate_summary(self):
        """ """
        self.logger.info(f"Saving to: {self.summary_path}")

        with PdfPages(self.summary_path) as pdf:
            for (name, alert) in tqdm(sorted(self.cache.items())):

                add_cutouts([alert])
                fig, _ = lightcurve_from_alert([alert], include_cutouts=True)

                pdf.savefig()
                plt.close()

    @staticmethod
    def parse_ztf_filter(fid: int):
        """ """
        return ["g", "r", "i"][fid - 1]

    def tns_summary(self):
        """ """
        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]
            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]
            print(
                "Candidate:",
                name,
                res["candidate"]["ra"],
                res["candidate"]["dec"],
                first_detection["jd"],
            )
            try:
                last_upper_limit = [
                    x
                    for x in res["prv_candidates"]
                    if np.logical_and(
                        "isdiffpos" in x.keys(), x["jd"] < first_detection["jd"]
                    )
                ][-1]
                print(
                    "Last Upper Limit:",
                    last_upper_limit["jd"],
                    self.parse_ztf_filter(last_upper_limit["fid"]),
                    last_upper_limit["diffmaglim"],
                )
            except IndexError:
                last_upper_limit = None
                print("Last Upper Limit: None")
            print(
                "First Detection:",
                first_detection["jd"],
                self.parse_ztf_filter(first_detection["fid"]),
                first_detection["magpsf"],
                first_detection["sigmapsf"],
            )
            hours_after_merger = 24.0 * (first_detection["jd"] - self.t_min.jd)
            print(f"First observed {hours_after_merger} hours after merger")
            if last_upper_limit:
                print(
                    "It has risen",
                    -latest["magpsf"] + last_upper_limit["diffmaglim"],
                    self.parse_ztf_filter(latest["fid"]),
                    self.parse_ztf_filter(last_upper_limit["fid"]),
                )
            print(
                [
                    x["jd"]
                    for x in res["prv_candidates"] + [res["candidate"]]
                    if "isdiffpos" in x.keys()
                ]
            )
            print("\n")

    def peak_mag_summary(self):
        for name, res in sorted(self.cache.items()):

            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_mags = [x["magpsf"] for x in detections]
            brightest = detections[detection_mags.index(min(detection_mags))]

            tns_result = ""
            tns_name, tns_date, tns_group = ampel_api_tns(
                brightest["ra"], brightest["dec"], searchradius_arcsec=3.0
            )
            if tns_name:
                tns_result = f"({tns_name})"

            xmatch_info = get_cross_match_info(raw=res, logger=self.logger)

            print(
                f"Candidate {name} peaked at {brightest['magpsf']:.1f} {tns_result}on "
                f"{brightest['jd']:.1f} with filter {self.parse_ztf_filter(brightest['fid'])} "
                f"{xmatch_info}"
            )

    def candidate_text(self, name, first_detection, lul_lim, lul_jd):
        raise NotImplementedError

    def text_summary(self):
        """ """
        text = ""
        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]
            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]
            try:
                last_upper_limit = [
                    x
                    for x in res["prv_candidates"]
                    if np.logical_and(
                        "isdiffpos" in x.keys(), x["jd"] < first_detection["jd"]
                    )
                ][-1]

                text += self.candidate_text(
                    name,
                    first_detection["jd"],
                    last_upper_limit["diffmaglim"],
                    last_upper_limit["jd"],
                )

            # No pre-detection upper limit
            except IndexError:
                text += self.candidate_text(name, first_detection["jd"], None, None)

            ned_z, ned_dist = query_ned_for_z(
                ra_deg=latest["ra"],
                dec_deg=latest["dec"],
                searchradius_arcsec=20,
                logger=self.logger,
            )

            if ned_z:
                ned_z = float(ned_z)
                absmag = self.calculate_abs_mag(latest["magpsf"], ned_z)
                if ned_z > 0:
                    z_dist = Distance(z=ned_z, cosmology=cosmo).value
                    text += f"It has a spec-z of {ned_z:.3f} [{z_dist:.0f} Mpc] and an abs. mag of {absmag:.1f}. Distance to SDSS galaxy is {ned_dist:.2f} arcsec. "
                    if self.dist:
                        gw_dist_interval = [
                            self.dist - self.dist_unc,
                            self.dist + self.dist_unc,
                        ]

            c = SkyCoord(res["candidate"]["ra"], res["candidate"]["dec"], unit="deg")
            g_lat = c.galactic.b.degree
            if abs(g_lat) < 15.0:
                text += f"It is located at a galactic latitude of {g_lat:.2f} degrees. "

            xmatch_info = get_cross_match_info(raw=res, logger=self.logger)
            text += xmatch_info
            text += "\n"
        return text

    def plot_overlap_with_observations(
        self, fields=None, pid=None, first_det_window_days=None, min_sep=0.01
    ):
        """ """

        nside = self.nside

        fig = plt.figure()
        plt.subplot(projection="aitoff")

        double_in_plane_probs = []
        single_in_plane_prob = []

        if fields is None:
            mns = self.get_multi_night_summary(first_det_window_days)

        else:

            class MNS:
                def __init__(self, data):
                    self.data = pandas.DataFrame(
                        data, columns=["field", "ra", "dec", "datetime"]
                    )

            data = []

            for f in fields:
                ra, dec = ztfquery_fields.get_field_centroid(f)[0]
                for i in range(2):
                    t = Time(self.t_min.jd + 0.1 * i, format="jd").utc
                    t.format = "isot"
                    t = t.value
                    data.append([f, ra, dec, t])

            mns = MNS(data)

        # Skip all 64 simultaneous quadrant entries, we only need one per observation for qa log
        # data = mns.data.copy().iloc[::64]

        data = mns.data.copy()

        ras = np.ones_like(data["field"]) * np.nan
        decs = np.ones_like(data["field"]) * np.nan

        # Actually load up ra/dec

        veto_fields = []

        for field in list(set(data["field"])):

            mask = data["field"] == field

            res = ztfquery_fields.get_field_centroid(field)

            if len(res) > 0:

                ras[mask] = res[0][0]
                decs[mask] = res[0][1]

            else:
                veto_fields.append(field)

        if len(veto_fields) > 0:
            self.logger.info(
                f"No RA/Dec found by ztfquery for fields {veto_fields}. "
                f"These observation have to be ignored."
            )

        data["ra"] = ras
        data["dec"] = decs

        mask = np.array([~np.isnan(x) for x in data["ra"]])

        data = data[mask]

        if pid is not None:
            pid_mask = data["pid"] == str(pid)
            data = data[pid_mask]

        obs_times = np.array(
            [
                Time(
                    data["datetime"].iat[i].replace(" ", "T"),
                    format="isot",
                    scale="utc",
                )
                for i in range(len(data))
            ]
        )

        if first_det_window_days is not None:
            first_det_mask = [
                x < Time(self.t_min.jd + first_det_window_days, format="jd").utc
                for x in obs_times
            ]
            data = data[first_det_mask]
            obs_times = obs_times[first_det_mask]

        pix_obs_times = dict()

        self.logger.info(f"Most recent observation found is {obs_times[-1]}")

        self.logger.info("Unpacking observations")
        pix_map = dict()

        for i, obs_time in enumerate(tqdm(obs_times)):

            # radec = [f"{data['ra'].iat[i]} {data['dec'].iat[i]}"]
            #
            # coords = SkyCoord(radec, unit=(u.deg, u.deg))
            pix = get_quadrant_ipix(nside, data["ra"].iat[i], data["dec"].iat[i])

            field = data["field"].iat[i]

            flat_pix = []

            for sub_list in pix:
                for p in sub_list:
                    flat_pix.append(p)

            flat_pix = list(set(flat_pix))

            t = obs_time.jd
            for p in flat_pix:
                if p not in pix_obs_times.keys():
                    pix_obs_times[p] = [t]
                else:
                    pix_obs_times[p] += [t]

                if p not in pix_map.keys():
                    pix_map[p] = [field]
                else:
                    pix_map[p] += [field]

        npix = hp.nside2npix(nside)
        theta, phi = hp.pix2ang(nside, np.arange(npix), nest=False)
        radecs = SkyCoord(ra=phi * u.rad, dec=(0.5 * np.pi - theta) * u.rad)
        idx = np.where(np.abs(radecs.galactic.b.deg) <= 10.0)[0]

        double_in_plane_pixels = []
        double_in_plane_probs = []
        single_in_plane_pixels = []
        single_in_plane_prob = []
        veto_pixels = []
        plane_pixels = []
        plane_probs = []
        times = []
        double_no_plane_prob = []
        double_no_plane_pixels = []
        single_no_plane_prob = []
        single_no_plane_pixels = []

        overlapping_fields = []
        for i, p in enumerate(tqdm(hp.nest2ring(nside, self.pixel_nos))):

            if p in pix_obs_times.keys():

                if p in idx:
                    plane_pixels.append(p)
                    plane_probs.append(self.map_probs[i])

                obs = pix_obs_times[p]

                # check which healpix are observed twice
                if max(obs) - min(obs) > min_sep:
                    # is it in galactic plane or not?
                    if p not in idx:
                        double_no_plane_prob.append(self.map_probs[i])
                        double_no_plane_pixels.append(p)
                    else:
                        double_in_plane_probs.append(self.map_probs[i])
                        double_in_plane_pixels.append(p)

                else:
                    if p not in idx:
                        single_no_plane_pixels.append(p)
                        single_no_plane_prob.append(self.map_probs[i])
                    else:
                        single_in_plane_prob.append(self.map_probs[i])
                        single_in_plane_pixels.append(p)

                overlapping_fields += pix_map[p]

                times += list(obs)
            else:
                veto_pixels.append(p)

        overlapping_fields = sorted(list(set(overlapping_fields)))
        self.overlap_fields = list(set(overlapping_fields))

        self.overlap_prob = np.sum(double_in_plane_probs + double_no_plane_prob) * 100.0

        size = hp.max_pixrad(nside) ** 2 * 50.0

        veto_pos = np.array(
            [hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in veto_pixels]
        ).T

        if len(veto_pos) > 0:

            plt.scatter(
                self.wrap_around_180(np.radians(veto_pos[0])),
                np.radians(veto_pos[1]),
                color="red",
                s=size,
            )

        plane_pos = np.array(
            [hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in plane_pixels]
        ).T

        if len(plane_pos) > 0:

            plt.scatter(
                self.wrap_around_180(np.radians(plane_pos[0])),
                np.radians(plane_pos[1]),
                color="green",
                s=size,
            )

        single_pos = np.array(
            [
                hp.pixelfunc.pix2ang(nside, i, lonlat=True)
                for i in single_no_plane_pixels
            ]
        ).T

        if len(single_pos) > 0:
            plt.scatter(
                self.wrap_around_180(np.radians(single_pos[0])),
                np.radians(single_pos[1]),
                c=single_no_plane_prob,
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
                cmap="gray",
            )

        plot_pos = np.array(
            [
                hp.pixelfunc.pix2ang(nside, i, lonlat=True)
                for i in double_no_plane_pixels
            ]
        ).T

        if len(plot_pos) > 0:
            plt.scatter(
                self.wrap_around_180(np.radians(plot_pos[0])),
                np.radians(plot_pos[1]),
                c=double_no_plane_prob,
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
            )

        red_patch = mpatches.Patch(color="red", label="Not observed")
        gray_patch = mpatches.Patch(color="gray", label="Observed once")
        violet_patch = mpatches.Patch(
            color="green", label="Observed Galactic Plane (|b|<10)"
        )
        plt.legend(handles=[red_patch, gray_patch, violet_patch])

        message = (
            "In total, {0:.2f} % of the contour was observed at least once. \n "
            "This estimate includes {1:.2f} % of the contour "
            "at a galactic latitude <10 deg. \n "
            "In total, {2:.2f} % of the contour was observed at least twice. \n"
            "In total, {3:.2f} % of the contour was observed at least twice, "
            "and excluding low galactic latitudes. \n"
            "These estimates account for chip gaps.".format(
                100
                * (
                    np.sum(double_in_plane_probs)
                    + np.sum(single_in_plane_prob)
                    + np.sum(single_no_plane_prob)
                    + np.sum(double_no_plane_prob)
                ),
                100 * np.sum(plane_probs),
                100.0 * (np.sum(double_in_plane_probs) + np.sum(double_no_plane_prob)),
                100.0 * np.sum(double_no_plane_prob),
            )
        )

        all_pix = single_in_plane_pixels + double_in_plane_pixels

        n_pixels = len(
            single_in_plane_pixels
            + double_in_plane_pixels
            + double_no_plane_pixels
            + single_no_plane_pixels
        )
        n_double = len(double_no_plane_pixels + double_in_plane_pixels)
        n_plane = len(plane_pixels)

        self.area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_pixels
        self.double_extragalactic_area = (
            hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_double
        )
        plane_area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_plane

        try:
            self.first_obs = Time(min(times), format="jd")
            self.first_obs.utc.format = "isot"
            self.last_obs = Time(max(times), format="jd")
            self.last_obs.utc.format = "isot"

        except ValueError:
            raise Exception(
                f"No observations of this field were found at any time between {self.t_min} and"
                f"{obs_times[-1]}. Coverage overlap is 0%, but recent observations might be missing!"
            )

        self.logger.info(f"Observations started at {self.first_obs.jd}")

        self.overlap_fields = overlapping_fields

        #     area = (2. * base_ztf_rad)**2 * float(len(overlapping_fields))
        #     n_fields = len(overlapping_fields)

        self.logger.info(
            f"{n_pixels} pixels were covered, covering approximately {self.area:.2g} sq deg."
        )
        self.logger.info(
            f"{n_double} pixels were covered at least twice (b>10), covering approximately {self.double_extragalactic_area:.2g} sq deg."
        )
        self.logger.info(
            f"{n_plane} pixels were covered at low galactic latitude, covering approximately {plane_area:.2g} sq deg."
        )
        return fig, message

    def crosscheck_prob(self):

        try:
            nside = self.ligo_nside
        except AttributeError:
            nside = self.nside

        class MNS:
            def __init__(self, data):
                self.data = pandas.DataFrame(
                    data, columns=["field", "ra", "dec", "datetime"]
                )

        data = []

        for f in self.overlap_fields:
            ra, dec = ztfquery_fields.field_to_coords(float(f))[0]
            t = Time(self.t_min.jd, format="jd").utc
            t.format = "isot"
            t = t.value
            data.append([f, ra, dec, t])

            mns = MNS(data)

        data = mns.data.copy()

        self.logger.info("Unpacking observations")
        field_prob = 0.0

        ps = []

        for index, row in tqdm(data.iterrows()):
            pix = get_quadrant_ipix(nside, row["ra"], row["dec"])

            flat_pix = []

            for sub_list in pix:
                for p in sub_list:
                    flat_pix.append(p)

            flat_pix = list(set(flat_pix))
            ps += flat_pix

        ps = list(set(ps))

        for p in hp.ring2nest(nside, ps):
            field_prob += self.data[self.key][int(p)]

        self.logger.info(
            f"Intergrating all fields overlapping 90% contour gives {100*field_prob:.2g}%"
        )



import datetime
import os
from astropy.time import Time
from ztfquery import skyvision
from ztfquery.io import LOCALSOURCE


def get_obs_summary(t_min, max_days=None):
    mns_time = str(t_min).split("T")[0].replace("-", "")
    now = datetime.datetime.now()

    if max_days is not None:
        date_1 = datetime.datetime.strptime(mns_time, "%Y%m%d")
        end_date = date_1 + datetime.timedelta(days=max_days)
        end_date = end_date.strftime("%Y-%m-%d")
    else:
        end_date = now.strftime("%Y-%m-%d")

    start_date_jd = t_min.jd
    start_date_jd = Time(start_date_jd, format="jd").jd

    # ztfquery saves nightly observations in a cache, and does not redownload them.
    # If the nightly log was not complete, it will never be updated.
    # Here we simply clear the cache and cleanly re-download everything.

    skyvision_log = os.path.join(LOCALSOURCE, "skyvision")

    for filename in os.listdir(skyvision_log):
        if ".csv" in filename:
            path = os.path.join(skyvision_log, filename)
            os.remove(path)

    mns = skyvision.CompletedLog.from_daterange(mns_time, end=end_date, verbose=False)

    mns.data["obsjd"] = Time(list(mns.data.datetime.values), format="isot").jd

    mns.data.query(f"obsjd > {start_date_jd}", inplace=True)

    mns.data.reset_index(inplace=True)
    mns.data.drop(columns=["index"], inplace=True)

    return mns



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import os, argparse, logging, re, getpass
from ampel.ztf.archive.ArchiveDB import ArchiveDB
import ampel.contrib.hu.t3.TNSTalker as TNSTalker
from ampel.ztf.utils.ZIAlertUtils import ZIAlertUtils
from ampel.ztf.pipeline.common.ZTFUtils import ZTFUtils
from ampel.contrib.hu.t3.ampel_tns import sendTNSreports

# Create a logger
logger = logging.getLogger(__name__)

# Run config for TNSTalker
ztf_default_values = {"flux_units": "1", "instrument_value": "196", "Observer": "Robot"}

RUN_CONFIG = {
    "submit_tns": True,
    "sandbox": True,
    "resubmit_tns_nonztf": False,
    "resubmit_tns_ztf": False,
    "submit_unless_journal": True,
    "lc_filters": [],
    "ztf_tns_at": ztf_default_values,
}

from nuztf.ampel_magic import username
from nuztf.ampel_magic import password

TNS_API_KEYFILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), ".TNS_api_key.cred"
)
try:
    with open(TNS_API_KEYFILE, "r") as f:
        TNS_API_KEY = f.read()
except FileNotFoundError:
    TNS_API_KEY = getpass.getpass(prompt="TNS API Key: ", stream=None)
    with open(TNS_API_KEYFILE, "wb") as f:
        f.write(TNS_API_KEY.encode())

RUN_CONFIG["max_age"] = 5
RUN_CONFIG["tns_api_key"] = TNS_API_KEY

# Connect to TNSTalker instance
TNS = TNSTalker.TNSTalker(logger=logger, run_config=RUN_CONFIG)
TNS.run_config = TNS.RunConfig(**RUN_CONFIG)
PORT = 5432

# Connect to AMPEL instance
AMPEL_CLIENT = ArchiveDB(
    "postgresql://{0}:{1}@localhost:{2}/ztfarchive".format(username, password, PORT)
)


def is_ztf_name(name):
    """ """
    return re.match("^ZTF[1-2]\d[a-z]{7}$", name)


def use_if_ztf(file_or_name):
    """ """
    if is_ztf_name(file_or_name):
        object_list = [file_or_name]
    else:
        object_list = []
        try:
            file = open(f"{file_or_name}", "r")
            lines = file.read().splitlines()
            for line in lines:
                if is_ztf_name(line):
                    object_list.append(line)
        except FileNotFoundError as error:
            print(
                "\nYou have to provide either a ZTF name or a file containing ZTF names (1 per line).\n"
            )
            raise error
        assert (
            object_list[0][:3] == "ZTF" and len(object_list[0]) == 12
        ), "You have to provide either a ZTF name or a file containing ZTF names (1 per line)"

    return object_list


def TNS_submit(ztf_names, reporter=None, sandbox=True):
    """
    Note: IF YOU WANT TO SUBMIT, SET sandbox=False
    """
    atreports = {}

    for ztf_name in ztf_names:
        ztf_object = AMPEL_CLIENT.get_alerts_for_object(ztf_name, with_history=True)
        query_res = [i for i in ztf_object]
        last_alert = query_res[-1]
        tv = ZIAlertUtils.to_transientview(
            file_path=None, content=last_alert, science_records=None
        )
        ztf_id = tv.tran_id
        ampel_id = ZTFUtils.to_ampel_id(ztf_id)
        lc_id = tv.lightcurves[0].id
        object.__setattr__(tv, "tran_id", ampel_id)
        object.__setattr__(tv, "latest_state", lc_id)
        atreport = TNS.create_atreport(tv)
        if reporter is None:
            atreport["reporter"] = "The Zwicky Transient Facility (ZTF) Collaboration"
        else:
            atreport[
                "reporter"
            ] = f"{reporter} for the Zwicky Transient Facility (ZTF) Collaboration"
        atreport["exptime"] = "300"
        atreports.update({ztf_name: atreport})

    # Reformat the report (code by AMPEL!)
    atreportlist = []
    atreport = {}
    k = 0
    for tranid in atreports.keys():
        if len(atreport) > 90:
            self.logger.info("adding another report to TNS submit")
            atreportlist.append({"at_report": atreport})
            atreport = {}
            k = 0
        atreport[int(k)] = atreports[tranid]
        k += 1
    atreportlist.append({"at_report": atreport})
    tnsreplies = sendTNSreports(
        atreportlist, TNS.run_config.tns_api_key, logger, sandbox=sandbox
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Submit ZTF object(s) to TNS")
    parser.add_argument(
        "name",
        type=str,
        help='Provide a ZTF name (e.g. "ZTF19aaelulu") or a .txt-file containing a list of ZTF names',
    )
    parser.add_argument(
        "--commit",
        "-commit",
        action="store_false",
        help="Run an actual commit (instead of just using the TNS sandbox",
    )
    parser.add_argument(
        "--reporter",
        "-reporter",
        type=str,
        default=None,
        help="Provide an author for the submission",
    )

    commandline_args = parser.parse_args()
    name = commandline_args.name
    sandbox = commandline_args.commit
    reporter = commandline_args.reporter
    object_list = use_if_ztf(name)

    if sandbox:
        logger.info(
            f"Submitting {object_list} to the TNS - SANDBOX MODE ONLY (to change, use -commit)"
        )
    else:
        logger.info(f"Submitting {object_list} to the TNS")

    TNS_submit(object_list, reporter, sandbox=sandbox)



from astropy.coordinates import SkyCoord
from astropy import units as u
from nuztf.ampel_api import ampel_api_catalog, ampel_api_name


def query_ned_for_z(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 20, logger=None
):

    z = None
    dist_arcsec = None

    query = ampel_api_catalog(
        catalog="NEDz_extcats",
        catalog_type="extcats",
        ra_deg=ra_deg,
        dec_deg=dec_deg,
        searchradius_arcsec=searchradius_arcsec,
        searchtype="nearest",
        logger=logger,
    )

    if query:
        z = query["body"]["z"]
        dist_arcsec = query["dist_arcsec"]
    return z, dist_arcsec


def ampel_api_tns(
    ra_deg: float, dec_deg: float, searchradius_arcsec: float = 3, logger=None
):
    """Function to query TNS via ampel api"""

    full_name = None
    discovery_date = None
    source_group = None

    res = ampel_api_catalog(
        catalog="TNS",
        catalog_type="extcats",
        ra_deg=ra_deg,
        dec_deg=dec_deg,
        searchradius_arcsec=searchradius_arcsec,
        searchtype="nearest",
        logger=logger,
    )

    if res:
        response_body = res["body"]
        name = response_body["objname"]
        prefix = response_body["name_prefix"]
        full_name = prefix + name
        discovery_date = response_body["discoverydate"]
        if "source_group" in response_body.keys():
            source_group = response_body["source_group"]["group_name"]

    return full_name, discovery_date, source_group


search_rad = 1.5


def get_cross_match_info(raw: dict, logger=None):
    """ """
    alert = raw["candidate"]

    label = ""

    # Check if known variable star (https://arxiv.org/pdf/1405.4290.pdf)

    if label == "":
        res = ampel_api_catalog(
            catalog="CRTS_DR1",
            catalog_type="extcats",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=5.0,
            logger=logger,
        )
        if res is not None:
            if logger:
                logger.info(res)
            label = f"[CRTS variable star: {res[0]['body']['name']} ({res[0]['dist_arcsec']:.2f} arsec)]"

    # Check if known QSO/AGN

    res = ampel_api_catalog(
        catalog="milliquas",
        catalog_type="extcats",
        ra_deg=alert["ra"],
        dec_deg=alert["dec"],
        searchradius_arcsec=search_rad,
        logger=logger,
    )
    if res is not None:
        if len(res) == 1:

            if "q" in res[0]["body"]["broad_type"]:
                label = f"[MILLIQUAS: {res[0]['body']['name']} - Likely QSO (prob = {res[0]['body']['qso_prob']}%) ({res[0]['dist_arcsec']:.2f} arsec)]"
            else:
                label = f"[MILLIQUAS: {res[0]['body']['name']} - '{res[0]['body']['broad_type']}'-type source ({res[0]['dist_arcsec']:.2f} arsec)]"
        else:
            label = "[MULTIPLE MILLIQUAS MATCHES]"

    # Check if measured parallax in Gaia (i.e galactic)

    if label == "":
        res = ampel_api_catalog(
            catalog="GAIADR2",
            catalog_type="catsHTM",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=5.0,
            logger=logger,
        )
        if res is not None:
            if res[0]["body"]["Plx"] is not None:
                plx_sig = res[0]["body"]["Plx"] / res[0]["body"]["ErrPlx"]
                if plx_sig > 3.0:
                    label = f"[GAIADR2: {plx_sig:.1f}-sigma parallax ({res[0]['dist_arcsec']:.2f} arsec)]"

    # Check if classified as probable star in SDSS

    if label == "":
        res = ampel_api_catalog(
            catalog="SDSSDR10",
            catalog_type="catsHTM",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=search_rad,
            logger=logger,
        )
        if res is not None:
            if len(res) == 1:
                if float(res[0]["body"]["type"]) == 6.0:
                    label = f"[SDSS Morphology: 'Star'-type source ({res[0]['dist_arcsec']:.2f} arsec)]"
            else:
                label = "[MULTIPLE SDSS MATCHES]"

    # WISE colour cuts (https://iopscience.iop.org/article/10.3847/1538-4365/)

    if label == "":
        res = ampel_api_catalog(
            catalog="wise_color",
            catalog_type="extcats",
            ra_deg=alert["ra"],
            dec_deg=alert["dec"],
            searchradius_arcsec=search_rad,
            logger=logger,
        )
        if res is not None:
            if len(res) == 1:
                w1mw2 = res[0]["body"]["W1mW2"]
                if w1mw2 > 0.8:
                    label = (
                        f"[Probable WISE-selected quasar:W1-W2={w1mw2:.1f}>0.8  "
                        f"({res[0]['dist_arcsec']:.2f} arsec)]"
                    )
                elif w1mw2 > 0.8:
                    label = (
                        f"[Possible WISE-selected quasar:W1-W2={w1mw2:.1f}>0.5  "
                        f"({res[0]['dist_arcsec']:.2f} arsec)]"
                    )
            else:
                label = "[MULTIPLE WISE MATCHES]"

    return label


def check_cross_match_info_by_name(name: str, logger=None):
    """ """
    return get_cross_match_info(
        raw=ampel_api_name(name, with_history=False, logger=logger)[0], logger=logger
    )



#!/usr/bin/env python
# coding: utf-8

import os, datetime, wget, logging, time, json

from tqdm import tqdm
import requests
from numpy.lib.recfunctions import append_fields
import lxml.etree
import fitsio
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path

from astropy_healpix import HEALPix
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.time import Time

import healpy as hp
from ztfquery.io import LOCALSOURCE

from nuztf.base_scanner import BaseScanner
from nuztf.ampel_api import ampel_api_healpix, ampel_api_name

BASE_LIGO_DIR = os.path.join(LOCALSOURCE, "LIGO_skymaps")
BASE_GRB_DIR = os.path.join(LOCALSOURCE, "GRB_skymaps")
LIGO_CANDIDATE_OUTPUT_DIR = os.path.join(LOCALSOURCE, "LIGO_candidates")
GRB_CANDIDATE_OUTPUT_DIR = os.path.join(LOCALSOURCE, "GRB_candidates")
LIGO_CANDIDATE_CACHE = os.path.join(LOCALSOURCE, "LIGO_cache")
GRB_CANDIDATE_CACHE = os.path.join(LOCALSOURCE, "GRB_cache")

for entry in [
    BASE_LIGO_DIR,
    BASE_GRB_DIR,
    LIGO_CANDIDATE_OUTPUT_DIR,
    GRB_CANDIDATE_OUTPUT_DIR,
    LIGO_CANDIDATE_CACHE,
    GRB_CANDIDATE_CACHE,
]:
    if not os.path.exists(entry):
        os.makedirs(entry)

GW_RUN_CONFIG = {
    "min_ndet": 1,  # Default:2
    "min_tspan": -1,  # Default 0, but that rejects everything!
    "max_tspan": 365,
    "min_rb": 0.3,
    "max_fwhm": 5.5,
    "max_elong": 1.4,
    "max_magdiff": 1.0,
    "max_nbad": 2,
    "min_sso_dist": 20,
    "min_gal_lat": 0.0,  # Default: 14
    "gaia_rs": 10.0,
    "gaia_pm_signif": 3,
    "gaia_plx_signif": 3,
    "gaia_veto_gmag_min": 9,
    "gaia_veto_gmag_max": 20,
    "gaia_excessnoise_sig_max": 999,
    "ps1_sgveto_rad": 1.0,
    "ps1_sgveto_th": 0.8,
    "ps1_confusion_rad": 3.0,
    "ps1_confusion_sg_tol": 0.1,
}


class RetractionError(Exception):
    """Base class for retracted event"""

    pass


class SkymapScanner(BaseScanner):
    def __init__(
        self,
        event_name: str = None,
        skymap_file: str = None,
        scan_mode: str = "grb",
        rev: int = None,
        prob_threshold: float = 0.9,
        cone_nside: int = 64,
        n_days: int = 3,
        logger=None,
    ):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

        if not scan_mode in ["gw", "grb"]:
            raise ValueError(f"Scan mode must be either 'gw' or 'grb'.")

        if cone_nside != 64:
            raise ValueError(
                "At the moment, nside for healpix must be 64 (the only value supported by AMPEL API)."
            )

        self.prob_threshold = prob_threshold
        self.n_days = n_days
        self.scan_mode = scan_mode

        if skymap_file is None and self.scan_mode == "gw":
            self.skymap_path, self.summary_path, self.event_name = self.get_gw_skymap(
                event_name=event_name, rev=rev
            )

        elif skymap_file is None and self.scan_mode == "grb":
            self.get_grb_skymap(event_name=event_name)

        else:
            basename = os.path.basename(skymap_file)

            if self.scan_mode == "gw":
                self.skymap_path = os.path.join(BASE_LIGO_DIR, basename)
            else:
                self.skymap_path = os.path.join(BASE_GRB_DIR, basename)

            if skymap_file[:8] == "https://" and scan_mode == "gw":
                self.logger.info(f"Downloading from: {skymap_file}")
                self.skymap_path = os.path.join(
                    BASE_LIGO_DIR, os.path.basename(skymap_file[7:])
                )
                wget.download(skymap_file, self.skymap_path)

            if self.scan_mode == "gw":
                self.summary_path = os.path.join(
                    LIGO_CANDIDATE_OUTPUT_DIR,
                    f"{os.path.basename(skymap_file)}_{self.prob_threshold}.pdf",
                )

            else:
                self.summary_path = os.path.join(
                    GRB_CANDIDATE_OUTPUT_DIR,
                    f"{os.path.basename(skymap_file)}_{self.prob_threshold}.pdf",
                )

            self.event_name = os.path.basename(skymap_file[7:])

        self.data, t_obs, self.hpm, self.key, self.dist, self.dist_unc = self.read_map()

        t_min = Time(t_obs, format="isot", scale="utc")

        self.logger.info(f"Event time: {t_min}")
        self.logger.info("Reading map")

        self.pixel_threshold = self.find_pixel_threshold(self.data[self.key])
        (
            self.map_coords,
            self.pixel_nos,
            self.map_probs,
            self.nside,
            self.pixel_area,
        ) = self.unpack_skymap()

        BaseScanner.__init__(
            self,
            run_config=GW_RUN_CONFIG,
            t_min=t_min,
            logger=logger,
            cone_nside=cone_nside,
        )

        # By default, accept things detected within 72 hours of event time
        self.default_t_max = Time(self.t_min.jd + self.n_days, format="jd")

        self.logger.info(f"Time-range is {self.t_min} -- {self.default_t_max.isot}")

    def get_alerts(self):
        """Scan the skymap area and get ZTF transients"""
        self.logger.info("Commencing skymap scan")

        alerts = []

        time_healpix_start = time.time()
        n_tot = 0
        self.queue = []

        for i, cone_id in enumerate(tqdm(list(self.cone_ids))):

            ra, dec = self.cone_coords[i]

            ipix = hp.ang2pix(64, ra, dec, nest=True, lonlat=True)

            self.logger.debug(
                f"API healpix search. Healpix = {cone_id}. Timespan = {self.default_t_max.jd-self.t_min.jd:.1f} days."
            )

            query_res = ampel_api_healpix(
                ipix=cone_id,
                t_min_jd=self.t_min.jd,
                t_max_jd=self.default_t_max.jd,
                logger=self.logger,
            )

            n_tot += len(query_res)
            self.queue.append((i, query_res))
            self.scanned_pixels.append(i)

        time_healpix_end = time.time()
        time_healpix = time_healpix_end - time_healpix_start

        if self.scan_mode == "gw":
            self.cache_dir = os.path.join(LIGO_CANDIDATE_CACHE, self.event_name)
        else:
            self.cache_dir = os.path.join(GRB_CANDIDATE_CACHE, self.event_name)

        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        cache_file = os.path.join(self.cache_dir, f"{self.event_name}_all_alerts.json")

        json.dump(self.queue, open(cache_file, "w"))

        self.logger.info(
            f"Added {n_tot} candidates detected between {self.t_min} and {self.default_t_max.isot}"
        )
        self.logger.info(f"This took {time_healpix:.1f} s in total")

    def filter_alerts(self, load_cachefile=False):
        """ """
        self.logger.info(f"Commencing first stage filtering.")
        cache_file = os.path.join(self.cache_dir, f"{self.event_name}_all_alerts.json")

        if load_cachefile:
            self.queue = json.load(open(cache_file, "r"))

        first_stage_objects = []
        filter_time = 0.0

        for i, item in enumerate(tqdm(self.queue)):

            (j, query_res) = item

            indices = []

            for i, res in enumerate(query_res):

                if self.filter_f_no_prv(
                    res=res,
                    t_min_jd=self.t_min.jd,
                    t_max_jd=self.default_t_max.jd,
                ):
                    indices.append(i)

            res = [query_res[i] for i in indices]

            first_stage_objects += [x["objectId"] for x in res]

        first_stage_objects = self.remove_duplicates(first_stage_objects)

        self.logger.info(
            f"First stage of filtering (based on predetections) took {filter_time:.1f} s in total. {len(first_stage_objects)} transients make the cut."
        )

        cache_file_first_stage = cache_file[:-15] + "_first_stage.json"
        json.dump(first_stage_objects, open(cache_file_first_stage, "w"))

        # Second and final stage
        self.logger.info(
            f"Second stage commencing: Now we do proper filtering based on PS1 etc."
        )

        start_secondfilter = time.time()

        final_objects = []

        for ztf_id in tqdm(first_stage_objects):

            query_res = ampel_api_name(ztf_name=ztf_id, with_history=True)

            for res in query_res:
                _ztf_id = res["objectId"]

                if self.filter_f_history(
                    res=res, t_min_jd=self.t_min.jd, t_max_jd=self.default_t_max.jd
                ):

                    if self.filter_ampel(res):
                        self.logger.debug(f"{_ztf_id}: Passed all cuts")
                        final_objects.append(_ztf_id)
                        self.cache[_ztf_id] = res
                    else:
                        self.logger.debug(f"{_ztf_id}: Rejected")
                else:
                    self.logger.debug(f"{_ztf_id}: Failed History")

        end_secondfilter = time.time()
        filter_time = end_secondfilter - start_secondfilter

        final_objects = self.remove_duplicates(final_objects)

        cache_file_final_stage = cache_file[:-15] + "_final_stage.json"

        json.dump(final_objects, open(cache_file_final_stage, "w"))

        self.logger.info(
            f"Final stage of filtering took {filter_time:.1f} s in total. {len(final_objects)} transients make the cut."
        )

    def remove_duplicates(self, ztf_ids: list):
        """ """
        return list(set(ztf_ids))

    def get_obs_line(self):
        """ """
        return "Each exposure was 30s with a typical depth of 20.5 mag."

    def get_overlap_line(self):
        """ """
        return (
            "We covered {0:.1f}% of the enclosed probability "
            "based on the map in {1:.1f} sq deg. "
            "This estimate accounts for chip gaps. ".format(
                self.overlap_prob, self.area
            )
        )

    @staticmethod
    def remove_variability_line():
        """ """
        return (
            ", and removing candidates with history of "
            "variability prior to the merger time"
        )

    def candidate_text(
        self, name: str, first_detection: float, lul_lim: float, lul_jd: float
    ):
        """ """
        try:
            text = (
                "{0}, first detected {1:.1f} hours after merger, "
                "was not detected {2:.1f} days prior to a depth of {3:.2f}. ".format(
                    name,
                    24.0 * (first_detection - self.t_min.jd),
                    first_detection - lul_jd,
                    lul_lim,
                )
            )
        except TypeError:
            text = (
                f"{name} had upper limit problems. PLEASE FILL IN NUMBERS BY HAND!!! "
            )

        return text

    def filter_f_no_prv(self, res: dict, t_min_jd: float, t_max_jd: float) -> bool:
        """First filtering stage"""

        # Veto transients older than t_min_jd
        # as we don't expect detections before GRB or GW event time)
        if res["candidate"]["jdstarthist"] < t_min_jd:
            startdate_jd = res["candidate"]["jdstarthist"]
            startdate_date = Time(startdate_jd, format="jd").isot
            self.logger.debug(
                f"{res['objectId']}: Transient is too old (jdstarthist predates event; first detection at {startdate_date})."
            )
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > t_max_jd:
            startdate_jd = res["candidate"]["jdstarthist"]
            startdate_date = Time(startdate_jd, format="jd").isot
            self.logger.debug(
                f"{res['objectId']}: Transient is too new (jdstarthist too late after event; first detection at {startdate_date})"
            )
            return False

        # Exclude negative detection
        if res["candidate"]["isdiffpos"] not in ["t", "1"]:
            self.logger.debug(f"{res['objectId']}: Negative subtraction")
            return False

        try:
            if res["candidate"]["drb"] < 0.3:
                self.logger.debug(f"{res['objectId']}: DRB too low")
                return False
        except (KeyError, TypeError):
            pass

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.debug(f"{res['objectId']}: Outside of event contour.")
            return False

        self.logger.debug(f"{res['objectId']}: Passed first filter stage (no prv)")

        return True

    def filter_f_history(self, res: dict, t_min_jd, t_max_jd):
        """Veto transients"""

        # Veto old transients
        ztf_id = res["objectId"]
        if res["candidate"]["jdstarthist"] < t_min_jd:
            self.logger.debug(
                f"{ztf_id}: Transient is too old. (jdstarthist history predates event)"
            )
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > t_max_jd:
            self.logger.debug(
                f"{ztf_id}: Transient is too new. (jdstarthist too late after event)"
            )
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            self.logger.debug(f"{ztf_id}: Not passed mover cut")
            return False

        # Require 2 positive detections
        old_detections = [
            x
            for x in res["prv_candidates"]
            if np.logical_and("isdiffpos" in x.keys(), x["jd"] > t_min_jd)
        ]

        pos_detections = [x for x in old_detections if "isdiffpos" in x.keys()]

        if len(pos_detections) < 1:
            self.logger.debug(f"{ztf_id}: Does not have two detections")
            return False

        self.logger.debug(f"{ztf_id}: Passed the filtering stage")

        return True

    def get_gw_skymap(self, event_name: str, rev: int):
        """ """
        from ligo.gracedb.rest import GraceDb

        ligo_client = GraceDb()

        self.logger.info("Superevent!")

        if event_name is None:
            superevent_iterator = ligo_client.superevents("category: Production")
            superevent_ids = [
                superevent["superevent_id"] for superevent in superevent_iterator
            ]
            event_name = superevent_ids[0]

        voevents = ligo_client.voevents(event_name).json()["voevents"]

        if rev is None:
            rev = len(voevents)

        elif rev > len(voevents):
            raise Exception("Revision {0} not found".format(rev))

        latest_voevent = voevents[rev - 1]
        self.logger.info(f"Found voevent {latest_voevent['filename']}")

        if "Retraction" in latest_voevent["filename"]:
            raise RetractionError(
                f"The specified LIGO event, {latest_voevent['filename']}, was retracted."
            )

        response = requests.get(latest_voevent["links"]["file"])

        root = lxml.etree.fromstring(response.content)
        params = {
            elem.attrib["name"]: elem.attrib["value"]
            for elem in root.iterfind(".//Param")
        }

        latest_skymap = params["skymap_fits"]

        self.logger.info(f"Latest skymap URL: {latest_skymap}")

        base_file_name = os.path.basename(latest_skymap)
        savepath = os.path.join(
            BASE_LIGO_DIR,
            f"{event_name}_{latest_voevent['N']}_{base_file_name}",
        )

        self.logger.info(f"Saving to: {savepath}")
        response = requests.get(latest_skymap)

        with open(savepath, "wb") as f:
            f.write(response.content)

        summary_path = f"{LIGO_CANDIDATE_OUTPUT_DIR}/{event_name}_{latest_voevent['N']}_{self.prob_threshold}.pdf"

        return savepath, summary_path, event_name

    def get_grb_skymap(self, event_name: str):
        """ """
        if event_name is None:
            raise ValueError(
                "event_name must be provided for GRBs. They must have the form 'GRB210729A"
            )

        event_year_short = event_name[3:5]
        event_year = "20" + event_year_short
        event_month = event_name[5:7]
        event_day = event_name[7:9]
        event_letter = event_name[9]
        event_number = ord(event_letter) - 65

        event_date = datetime.datetime(
            int(event_year), int(event_month), int(event_day)
        )

        # get possible skymap URLs

        url = f"https://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/triggers/{event_date.year}"

        # response = requests.get(url)
        # print(response.html)
        from lxml import html

        page_overview = requests.get(url)
        webpage_overview = html.fromstring(page_overview.content)

        links_overview = webpage_overview.xpath("//a/@href")

        links_for_date = []

        for link in links_overview:
            if link[2:8] == f"{event_year_short}{event_month}{event_day}":
                links_for_date.append(url + "/" + link + "current/")

        if len(links_for_date) > 1:
            self.logger.info(
                f"Found multiple events. Will choose the one corresponding the GRB letter {event_letter}"
            )

        event_url = links_for_date[event_number]

        page_event = requests.get(event_url)
        webpage_event = html.fromstring(page_event.content)
        links_event = webpage_event.xpath("//a/@href")

        for link in links_event:
            if link[0:11] == "glg_healpix":
                final_link = event_url + link
                break

        self.skymap_path = os.path.join(BASE_GRB_DIR, link)

        if os.path.isfile(self.skymap_path):
            self.logger.info(
                f"Continuing with saved skymap. Located at {self.skymap_path}"
            )
        else:
            self.logger.info(f"Downloading skymap and saving to {self.skymap_path}")
            wget.download(final_link, self.skymap_path)

        self.summary_path = (
            f"{GRB_CANDIDATE_OUTPUT_DIR}/{event_name}_{self.prob_threshold}.pdf"
        )

        self.event_name = event_name

    def read_map(
        self,
    ):
        """Read the skymap"""

        self.logger.info(f"Reading file: {self.skymap_path}")

        data, h = fitsio.read(self.skymap_path, header=True)

        if "DISTMEAN" not in h:
            dist = None
        else:
            dist = h["DISTMEAN"]
        if "DISTSTD" not in h:
            dist_unc = None
        else:
            dist_unc = h["DISTSTD"]
        if "DATE-OBS" not in h:
            t_obs = fitsio.read_header(self.skymap_path)["DATE-OBS"]
        else:
            t_obs = h["DATE-OBS"]

        if "PROB" in data.dtype.names:
            key = "PROB"
        elif "PROBABILITY" in data.dtype.names:
            key = "PROB"
            prob = np.array(data["PROBABILITY"]).flatten()
            data = append_fields(data, "PROB", prob)
        else:
            raise Exception(
                "No recognised probability key in map. This is probably a weird one, right?"
            )

        if not isinstance(data[0], float):
            probs = np.array(data["PROB"]).flatten()
            data = np.array(probs, dtype=np.dtype([("PROB", np.float)]))

        self.logger.info(f"Summed probability is {100. * np.sum(data['PROB']):.1f}%")

        if h["ORDERING"] == "RING":
            data["PROB"] = hp.pixelfunc.reorder(data["PROB"], inp="RING", out="NESTED")
            h["ORDERING"] = "NESTED"

        hpm = HEALPix(nside=h["NSIDE"], order=h["ORDERING"], frame="icrs")

        return data, t_obs, hpm, key, dist, dist_unc

    def find_pixel_threshold(self, data):
        """ """

        ranked_pixels = np.sort(data)[::-1]
        int_sum = 0.0
        pixel_threshold = 0.0

        for i, prob in enumerate(ranked_pixels):
            int_sum += prob
            if int_sum > self.prob_threshold:
                self.logger.info(
                    f"Threshold found! \n To reach {int_sum * 100.0}% of probability, pixels with probability greater than {prob} are included."
                )
                pixel_threshold = prob
                break

        return pixel_threshold

    def unpack_skymap(self):
        """ """

        nside = hp.npix2nside(len(self.data[self.key]))

        threshold = self.find_pixel_threshold(self.data[self.key])

        mask = self.data[self.key] > threshold

        map_coords = []

        pixel_nos = []

        self.logger.info("Checking which pixels are within the contour:")

        for i in tqdm(range(hp.nside2npix(nside))):
            if mask[i]:
                map_coords.append(self.extract_ra_dec(nside, i))
                pixel_nos.append(i)

        pixel_area = hp.nside2pixarea(nside, degrees=True) * float(len(map_coords))

        self.logger.info(f"Total pixel area: {pixel_area} degrees")

        map_coords = np.array(
            map_coords, dtype=np.dtype([("ra", np.float), ("dec", np.float)])
        )

        return map_coords, pixel_nos, self.data[self.key][mask], nside, pixel_area

    def find_cone_coords(self):
        """ """

        cone_ids = []

        for ra, dec in self.map_coords:
            cone_ids.append(self.extract_npix(self.cone_nside, ra, dec))

        cone_ids = list(set(cone_ids))

        cone_coords = []

        for i in tqdm(cone_ids):
            cone_coords.append(self.extract_ra_dec(self.cone_nside, i))

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", np.float), ("dec", np.float)])
        )

        return cone_ids, cone_coords

    def plot_skymap(self):
        """ """
        fig = plt.figure()
        plt.subplot(211, projection="aitoff")

        mask = self.data[self.key] > self.pixel_threshold

        size = hp.max_pixrad(self.nside, degrees=True) ** 2

        ra_map_rad = np.deg2rad(self.wrap_around_180(self.map_coords["ra"]))
        dec_map_rad = np.deg2rad(self.map_coords["dec"])

        ra_cone_rad = np.deg2rad(self.wrap_around_180(self.cone_coords["ra"]))
        dec_cone_rad = np.deg2rad(self.cone_coords["dec"])

        plt.scatter(
            ra_map_rad,
            dec_map_rad,
            c=self.data[self.key][mask],
            vmin=0.0,
            vmax=max(self.data[self.key]),
            s=size,
        )

        plt.title("SKYMAP")

        plt.subplot(212, projection="aitoff")

        plt.scatter(ra_cone_rad, dec_cone_rad)
        plt.title("CONE REGION")

        outpath = os.path.join(BASE_GRB_DIR, f"{self.event_name}.png")
        plt.tight_layout()

        plt.savefig(outpath, dpi=300)

        return fig

    def plot_coverage(self):
        """Plot ZTF coverage of skymap region"""
        fig, message = self.plot_overlap_with_observations(
            first_det_window_days=self.n_days
        )
        plt.tight_layout()

        outpath = os.path.join(
            GRB_CANDIDATE_OUTPUT_DIR, f"{self.event_name}_coverage.png"
        )
        plt.savefig(outpath, dpi=300)

        self.logger.info(message)

        return fig, message

    def interpolate_map(self, ra_deg, dec_deg):
        """ """
        interpol_map = self.hpm.interpolate_bilinear_skycoord(
            SkyCoord(ra_deg * u.deg, dec_deg * u.deg), self.data[self.key]
        )
        return interpol_map

    def in_contour(self, ra_deg, dec_deg):
        """ """
        return self.interpolate_map(ra_deg, dec_deg) > self.pixel_threshold


if __name__ == "__main__":

    import logging

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    scanner = SkymapScanner(logger=logger)
    scanner.plot_skymap()
    scanner.get_alerts()
    scanner.filter_alerts()



import numpy as np
import os
from slack import RTMClient
from slack_bot import bot_access_token, access_token

submit_file = os.path.join(
    os.path.dirname(os.path.abspath(__file__)) + "/spawn_tmux_session.sh"
)

ampel_bot_user = ["UMNJK00CU", "DMBKJG00K"]

keywords = ["<@{0}>".format(ampel_bot_user), "LIGO", "banana", "GRB", "Fermi", "Ampel"]


def run_on_event(data):
    print(data)
    ts = data["ts"]
    channel = data["channel"]
    sid = int(float(ts) * 1.0e6)
    cmd = "bash {0} {1} {2} {3}".format(submit_file, sid, ts, channel)
    print(cmd)
    os.system(cmd)


@RTMClient.run_on(event="message")
def say_hello(**payload):
    data = payload["data"]
    if "user" in data.keys():
        try:
            if not np.logical_and(
                np.sum([x in data["text"] for x in keywords]) == 0,
                "DMBKJG00K" not in data["user"],
            ):
                run_on_event(data)
        except KeyError:
            pass


if __name__ == "__main__":
    print("Running master client!")
    rtm_client = RTMClient(token=bot_access_token)
    rtm_client.start()



from slack import WebClient
import getpass
import numpy as np
import logging
import io
import os
from nuztf.gw_multi_process import MultiGwProcessor
import traceback
import time

slack_token = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), ".slack_access_token.txt"
)

try:
    with open(slack_token, "r") as f:
        access_token = f.read()
except FileNotFoundError:
    access_token = getpass.getpass(prompt="Slack Access Token: ", stream=None)
    with open(slack_token, "wb") as f:
        f.write(access_token.encode())

bot_token = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), ".slack_bot_access_token.txt"
)

try:
    with open(bot_token, "r") as f:
        bot_access_token = f.read()
except FileNotFoundError:
    bot_access_token = getpass.getpass(prompt="Slack Bot Access Token: ", stream=None)
    with open(bot_token, "wb") as f:
        f.write(bot_access_token.encode())


def upload_fig(fig, data, filename, channel_id, thread_ts):
    imgdata = io.BytesIO()
    fig.savefig(imgdata, format="png", dpi=600, transparent=True)
    imgdata.seek(0)
    wc = WebClient(token=bot_access_token)
    wc.files_upload(
        file=imgdata.getvalue(),
        filename=filename,
        channels=channel_id,
        thread_ts=thread_ts,
        icon_emoji=":ampel-mm:",
        text="<@{0}>, here's the file {1} I've uploaded for you!".format(
            data["user"], filename
        ),
    )
    # fig.close()


def run_on_event(thread_ts, channel_id):

    web_client = WebClient(token=access_token)

    payload = web_client.conversations_history(
        channel=channel_id,
        oldest=str(float(thread_ts) - 1),
        latest=str(float(thread_ts) + 1),
    )

    data = payload["messages"][0]

    user = data["user"]

    web_client.chat_postMessage(
        channel=channel_id,
        text="Hi <@{0}>! You are interested in Ampel multi-messenger stuff, right? Let me get right on that for you.".format(
            user
        ),
        thread_ts=thread_ts,
        icon_emoji=":ampel-mm:",
    )

    message = data["text"].replace(u"\xa0", u" ")

    split_message = message.split(" ")

    gw_name = None
    gw_file = None
    rev_no = None
    prob_threshold = 0.95
    n_days = None
    try:
        for x in split_message:
            if len(x) > 0:
                if x[0] in ["s", "S"]:
                    if np.sum([y.isdigit() for y in x[1:7]]) == 6:
                        gw_name = x
                elif ".fit" in x:
                    # print(gw_file)
                    # gw_file = x[1:-1]
                    gw_file = x[1:-1]
                elif "rev" in x:
                    rev_no = int(x.split("=")[1])
                elif "prob_threshold" in x:
                    prob_threshold = float(x.split("=")[1])
                elif "n_days" in x:
                    n_days = float(x.split("=")[1])

        message = ""

        if gw_name is not None:
            message = "You are interested in LIGO event {0}. ".format(gw_name)

        if gw_file is not None:
            if gw_name is not None:
                message = (
                    "You have also specified a fits file. The fits file will be used.  "
                )
            else:
                message = "You are interested in the following fits file: {0}. ".format(
                    gw_file
                )

        if message == "":
            message = "No file was specified. I will just assume that you want the most recent LIGO event. "

        if gw_file is None:

            if rev_no is not None:
                if gw_name is not None:
                    message += "You have specified revision number {0}. ".format(rev_no)
            else:
                message += "No revision number has been specified. I will just take the most recent revision for this event. "

        message += "The Skymap will be scanned up to {0}% of the probability. ".format(
            100.0 * prob_threshold
        )

        if n_days is None:
            message += "No time range has been specified. I will scan from merger time to now. "
        else:
            message += "I will scan for objects first detected between merger time and {0} days after merger. ".format(
                n_days
            )

    except:
        message = "Sorry <@{0}>, I have run into a parsing error with your message. All your bases are belong to us. \n {1}".format(
            data["channel"], split_message
        )

    web_client.chat_postMessage(
        channel=channel_id, text=message, thread_ts=thread_ts, icon_emoji=":ampel-mm:"
    )

    logger = logging.getLogger("quiet_logger")
    logger.setLevel(logging.INFO)

    try:
        gw = MultiGwProcessor(
            n_days=n_days,
            gw_name=gw_name,
            gw_file=gw_file,
            rev=rev_no,
            logger=logger,
            prob_threshold=prob_threshold,
        )
        web_client.chat_postMessage(
            channel=channel_id,
            text="Scanning method: {0} \n Effective sky number: {1}".format(
                gw.scan_method, gw.n_sky
            ),
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
        )
        fig = gw.plot_skymap()
        upload_fig(fig, data, "LIGO_skymap.png", channel_id, thread_ts)
        gw.clean_cache()
        gw.fill_queue()
        gw.terminate()
        gw.combine_cache()
        gw.clean_cache()
        wc = WebClient(token=bot_access_token)
        wc.files_upload(
            file=gw.output_path,
            filename=os.path.basename(gw.output_path),
            channels=channel_id,
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
        )
        fig, overlap = gw.plot_overlap_with_observations(first_det_window_days=n_days)
        web_client.chat_postMessage(
            channel=channel_id,
            text=overlap,
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
        )
        upload_fig(fig, data, "LIGO_overlap.png", channel_id, thread_ts)
        web_client.chat_postMessage(
            channel=channel_id,
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
            text="Here's a draft GCN:",
        )
        web_client.chat_postMessage(
            channel=channel_id,
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
            text=gw.draft_gcn(),
        )
        web_client.chat_postMessage(
            channel=channel_id,
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
            text="<@{0}>, I'm all finished. Go find that counterpart!".format(
                data["user"]
            ),
        )
    except Exception as e:
        web_client.chat_postMessage(
            channel=channel_id,
            text="Sorry <@{0}>, we need to talk. It's not you, it's me. I have run into an error, and cannot process your request further. I wish you the best of luck in all your future endeavours. \n\n `{1}`. ".format(
                data["user"], e
            ),
            thread_ts=thread_ts,
            icon_emoji=":ampel-mm:",
        )
        traceback.print_exc()
        time.sleep(120)
        raise

    # Session will not die until multi-processes have been terminated

    try:
        gw.terminate()
    except:
        pass
    print("Done!")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--timestamp", type=str)
    parser.add_argument("-c", "--channel", type=str)

    cfg = parser.parse_args()

    run_on_event(cfg.timestamp, cfg.channel)



import io, logging, gzip
import requests
import backoff
from base64 import b64decode
from json import JSONDecodeError
import numpy as np

from astropy.time import Time
from astropy.io import fits
from requests.auth import HTTPBasicAuth

from nuztf.credentials import load_credentials

# AMPEL API URLs

API_BASEURL = "https://ampel.zeuthen.desy.de"
API_ZTF_ARCHIVE_URL = API_BASEURL + "/api/ztf/archive/v2"
API_CATALOGMATCH_URL = API_BASEURL + "/api/catalogmatch"
API_CUTOUT_URL = API_BASEURL + "/api/ztf/archive/v2/cutouts"

_, ampel_api_archive_token = load_credentials("ampel_api_archive_token")


def merge_alerts(alert_list: list) -> list:
    """ """
    merged_list = []
    keys = list(set([x["objectId"] for x in alert_list]))

    for objectid in keys:
        alerts = [x for x in alert_list if x["objectId"] == objectid]
        if len(alerts) == 1:
            merged_list.append(alerts[0])
        else:
            jds = [x["candidate"]["jd"] for x in alerts]
            order = [jds.index(x) for x in sorted(jds)[::-1]]
            latest = alerts[jds.index(max(jds))]
            latest["candidate"]["jdstarthist"] = min(
                [x["candidate"]["jdstarthist"] for x in alerts]
            )

            for index in order[1:]:

                x = alerts[index]

                # Merge previous detections

                for prv in x["prv_candidates"] + [x["candidate"]]:
                    if prv not in latest["prv_candidates"]:
                        latest["prv_candidates"] = [prv] + latest["prv_candidates"]

            merged_list.append(latest)
    return merged_list


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_cone(
    ra: float,
    dec: float,
    radius: float,
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Function to query ampel via a cone search"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_conesearch = (
        API_ZTF_ARCHIVE_URL + f"/alerts/cone_search?ra={ra}&dec={dec}&"
        f"radius={radius}&jd_start={t_min_jd}&"
        f"jd_end={t_max_jd}&with_history={hist}&"
        f"with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_conesearch)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_conesearch,
        headers=headers,
    )
    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_timerange(
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Function to query ampel via a time-range search"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_timerange = (
        API_ZTF_ARCHIVE_URL + f"/alerts/time_range?jd_start={t_min_jd}&"
        f"jd_end={t_max_jd}&with_history={hist}&"
        f"with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_timerange)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_timerange,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return query_res


def add_cutouts(alert: list):
    candid = alert[0]["candid"]
    cutouts = ampel_api_cutout(candid)

    final_cutouts = {}

    if "detail" in cutouts.keys():
        if cutouts["detail"] == "Not Found":
            for k in ["science", "difference", "template"]:
                final_cutouts[f"cutout{k.title()}"] = {"data": create_empty_cutout()}
    else:
        for k in cutouts:
            final_cutouts[f"cutout{k.title()}"] = {
                "data": b64decode(cutouts[k]),
            }

    alert[0].update({"cutouts": final_cutouts})


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_name(
    ztf_name: str, with_history: bool = True, with_cutouts: bool = False, logger=None
) -> list:
    """Function to query ampel via name"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    queryurl_ztf_name = (
        API_ZTF_ARCHIVE_URL + f"/object/{ztf_name}/alerts?with_history={hist}"
    )

    logger.debug(queryurl_ztf_name)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_ztf_name,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()]
        query_res = merge_alerts(query_res)

    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    if with_cutouts:
        add_cutouts(query_res)

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_healpix(
    ipix: int,
    t_min_jd=Time("2018-04-01T00:00:00.123456789", format="isot", scale="utc").jd,
    t_max_jd=Time.now().jd,
    with_history: bool = False,
    with_cutouts: bool = False,
    chunk_size: int = 500,
    logger=None,
) -> list:
    """Function to query ampel based on a healpix pixel-index (gird has nside=64)"""

    if logger is None:
        logger = logging.getLogger(__name__)

    if with_history:
        hist = "true"
    else:
        hist = "false"

    if with_cutouts:
        cutouts = "true"
    else:
        cutouts = "false"

    queryurl_healpix = (
        API_ZTF_ARCHIVE_URL
        + f"/alerts/healpix?ipix={ipix}&jd_start={t_min_jd}&jd_end={t_max_jd}&with_history={hist}&with_cutouts={cutouts}&chunk_size={chunk_size}"
    )

    logger.debug(queryurl_healpix)

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_healpix,
        headers=headers,
    )

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        query_res = [i for i in response.json()["alerts"]]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_cutout(candid: int, logger=None):
    """Function to query ampel for cutouts by candidate ID"""

    if logger is None:
        logger = logging.getLogger(__name__)

    queryurl_cutouts = API_CUTOUT_URL + f"/{candid}"

    headers = {"Authorization": f"Bearer {ampel_api_archive_token}"}

    response = requests.get(
        queryurl_cutouts,
        headers=headers,
    )

    logger.debug(queryurl_cutouts)

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        cutouts = response.json()
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return cutouts


def create_empty_cutout():
    """Function to reate an empty image for missing cutouts"""
    npix = 63

    blank = np.ones((npix, npix))

    for i in range(npix):
        c = abs(npix / 2 - i) / (0.5 * npix)
        blank[i - 1][i - 1] = c
        blank[i - 1][npix - i - 1] = c

    hdu = fits.PrimaryHDU(blank)
    hdul = fits.HDUList([hdu])
    comp = io.BytesIO()
    hdul.writeto(comp)
    blank_compressed = gzip.compress(comp.getvalue())

    return blank_compressed


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def ampel_api_catalog(
    catalog: str,
    catalog_type: str,
    ra_deg: float,
    dec_deg: float,
    searchradius_arcsec: float = 10,
    searchtype: str = "all",
    logger=None,
):
    """
    Method for querying catalogs via the Ampel API
    'catalog' must be the name of a supported catalog, e.g.
    SDSS_spec, PS1, NEDz_extcats...
    For a full list of catalogs, confer
    https://ampel.zeuthen.desy.de/api/catalogmatch/catalogs

    """
    assert catalog_type in ["extcats", "catsHTM"]
    assert searchtype in ["all", "nearest"]

    if logger is None:
        logger = logging.getLogger(__name__)

    queryurl_catalogmatch = API_CATALOGMATCH_URL + "/cone_search/" + searchtype

    # First, we create a json body to post
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    query = {
        "ra_deg": ra_deg,
        "dec_deg": dec_deg,
        "catalogs": [
            {"name": catalog, "rs_arcsec": searchradius_arcsec, "use": catalog_type}
        ],
    }

    logger.debug(queryurl_catalogmatch)
    logger.debug(query)

    response = requests.post(url=queryurl_catalogmatch, json=query, headers=headers)

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    try:
        res = response.json()[0]
    except JSONDecodeError:
        if response.headers:
            logger.debug(response.headers)
        raise requests.exceptions.RequestException

    return res



import re, logging

import requests
import numpy as np
from astropy.time import Time

base_gcn_url = "https://gcn.gsfc.nasa.gov/gcn3"


def gcn_url(gcn_number):
    """ """
    return f"{base_gcn_url}/{gcn_number}.gcn3"


class ParsingError(Exception):
    """Base class for parsing error"""

    pass


def parse_gcn_archive():
    """ """
    page = requests.get(f"{base_gcn_url}_archive.html")

    nu_circulars = []

    for line in page.text.splitlines():
        if "IceCube observation of a high-energy neutrino" in line:
            res = line.split(">")
            gcn_no = "".join([x for x in res[2] if x.isdigit()])
            name = res[3].split(" - ")[0]
            nu_circulars.append((name, gcn_no))

    return nu_circulars


def parse_gcn_for_no(
    base_nu_name: str, url: str = f"{base_gcn_url}_archive.html", logger=None
):
    """ """
    if logger:
        logger.info(f"Checking for GCN on {url}")

    nu_name = str(base_nu_name)

    page = requests.get(url)

    gcn_no = None
    name = None

    while not nu_name[0].isdigit():
        nu_name = nu_name[1:]

    latest_archive_no = None

    for line in page.text.splitlines():
        if np.logical_and(
            "IceCube observation of a high-energy neutrino" in line, nu_name in line
        ):
            res = line.split(">")
            if gcn_no is None:
                gcn_no = "".join([x for x in res[2] if x.isdigit()])
                name = res[3].split(" - ")[0]
                print(f"Found match to {base_nu_name}: {name}")
            else:
                raise Exception(f"Multiple matches found to {base_nu_name}")

        elif np.logical_and("gcn3_arch_old" in line, latest_archive_no is None):
            url = line.split('"')[1]
            latest_archive_no = int(url[13:].split(".")[0])

    return gcn_no, name, latest_archive_no


def find_gcn_no(base_nu_name: str, logger=None):
    """ """
    gcn_no, name, latest_archive_no = parse_gcn_for_no(base_nu_name, logger=logger)

    if gcn_no is None:
        logging.info(
            f"No GCN found for {base_nu_name} on GCN page, checking archive instead. "
            f"The latest page is {latest_archive_no}"
        )

        while np.logical_and(latest_archive_no > 0, gcn_no is None):
            gcn_no, name, _ = parse_gcn_for_no(
                base_nu_name,
                url=f"{base_gcn_url}_arch_old{latest_archive_no}.html",
            )
            latest_archive_no -= 1

    # while

    if name is None:
        raise ParsingError("No GCN match found for {0}".format(base_nu_name))

    logging.info(f"Match is {name} (GCN #{gcn_no})")

    return gcn_no


def get_latest_gcn(logger=None):
    """ """
    latest = parse_gcn_archive()[0]
    if logger:
        logger.info(f"Latest GCN is {latest[0]} (GCN #{latest[1]})")
    return latest[1]


def parse_radec(str: str):
    """ """
    regex_findall = re.findall(r"[-+]?\d*\.\d+|\d+", str)
    if len(regex_findall) == 4:
        pos = float(regex_findall[0])
        pos_upper = float(regex_findall[1])
        pos_lower = float(regex_findall[1])
    elif len(regex_findall) == 5:
        pos, pos_upper, pos_lower = regex_findall[0:3]
        pos = float(pos)
        pos_upper = float(pos_upper.replace("+", ""))
        pos_lower = float(pos_lower.replace("-", ""))
    else:
        raise ParsingError(f"Could not parse GCN ra and dec")

    return pos, pos_upper, pos_lower


def parse_gcn_circular(gcn_number: int):
    """ """
    url = f"https://gcn.gsfc.nasa.gov/gcn3/{gcn_number}.gcn3"
    response = requests.get(url)
    returndict = {}
    mainbody_starts_here = 999
    splittext = response.text.splitlines()
    splittext = list(filter(None, splittext))
    for i, line in enumerate(splittext):
        if "SUBJECT" in line:
            name = line.split(" - ")[0].split(": ")[1]
            returndict.update({"name": name})
        elif "FROM" in line:
            base = line.split("at")[0].split(": ")[1].split(" ")
            author = [x for x in base if x != ""][1]
            returndict.update({"author": author})
        elif (
            ("RA" in line or "Ra" in line)
            and ("DEC" in splittext[i + 1] or "Dec" in splittext[i + 1])
            and i < mainbody_starts_here
        ):
            ra, ra_upper, ra_lower = parse_radec(line)
            dec, dec_upper, dec_lower = parse_radec(splittext[i + 1])
            ra_err = [ra_upper, -ra_lower]
            dec_err = [dec_upper, -dec_lower]
            returndict.update(
                {"ra": ra, "ra_err": ra_err, "dec": dec, "dec_err": dec_err}
            )
            mainbody_starts_here = i + 2
        elif ("Time" in line or "TIME" in line) and i < mainbody_starts_here:
            raw_time = [
                x for x in line.split(" ") if x not in ["Time", "", "UT", "UTC"]
            ][1]
            raw_time = "".join(
                [x for x in raw_time if np.logical_or(x.isdigit(), x in [":", "."])]
            )
            raw_date = name.split("-")[1][:6]
            ut_time = f"20{raw_date[0:2]}-{raw_date[2:4]}-{raw_date[4:6]}T{raw_time}"
            time = Time(ut_time, format="isot", scale="utc")
            returndict.update({"time": time})

    return returndict



#!/usr/bin/env python3
# coding: utf-8

import os, logging

from astropy.time import Time
import healpy as hp
import numpy as np
from tqdm import tqdm
from os import environ
from pathlib import Path
import requests

from ztfquery.io import LOCALSOURCE

from nuztf.base_scanner import BaseScanner
from nuztf.parse_nu_gcn import find_gcn_no, parse_gcn_circular, get_latest_gcn


nu_candidate_output_dir = os.path.join(LOCALSOURCE, "neutrino_candidates")

if not os.path.exists(nu_candidate_output_dir):
    os.makedirs(nu_candidate_output_dir)

NU_RUN_CONFIG = {
    "min_ndet": 1,  # Default:2
    "min_tspan": -1,  # Default 0, but that rejects everything!
    "max_tspan": 365,
    "min_rb": 0.0,
    "max_fwhm": 5.5,
    "max_elong": 1.4,
    "max_magdiff": 1.0,
    "max_nbad": 2,
    "min_sso_dist": 20,
    "min_gal_lat": -1.0,  # Default: 14
    "gaia_rs": 20,
    "gaia_pm_signif": 3,
    "gaia_plx_signif": 3,
    "gaia_veto_gmag_min": 9,
    "gaia_veto_gmag_max": 20,
    "gaia_excessnoise_sig_max": 999,
    "ps1_sgveto_rad": 1,
    "ps1_sgveto_th": 0.8,
    "ps1_confusion_rad": 3,
    "ps1_confusion_sg_tol": 0.1,
}


class NeutrinoScanner(BaseScanner):
    def __init__(
        self,
        nu_name: str = None,
        manual_args=None,
        gcn_no: int = None,
        cone_nside: int = 128,
        t_precursor: float = None,
        logger=None,
    ):

        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

        self.prob_threshold = 0.9

        if manual_args is None:

            if nu_name is not None:
                gcn_no = find_gcn_no(nu_name, logger=self.logger)

            elif gcn_no is None:
                gcn_no = get_latest_gcn(logger=self.logger)

            gcn_info = parse_gcn_circular(gcn_no)

            self.logger.info(gcn_info)

            nu_name = gcn_info["name"]
            author = gcn_info["author"]
            ra = [gcn_info["ra"], gcn_info["ra_err"][0], gcn_info["ra_err"][1]]
            dec = [gcn_info["dec"], gcn_info["dec_err"][0], gcn_info["dec_err"][1]]
            nu_time = gcn_info["time"]

            if t_precursor is not None:
                nu_time = Time(nu_time.mjd - t_precursor, format="mjd")

        else:
            (nu_name, ra, dec, nu_time) = manual_args
            author = None

        self.nu_name = nu_name
        self.author = author
        self.gcn_no = gcn_no
        self.dist = None

        self.logger.info(f"Neutrino time: {nu_time}")

        self.ra_max = float(max(ra[1:]) + ra[0])
        self.ra_min = float(min(ra[1:]) + ra[0])
        self.dec_max = float(max(dec[1:]) + dec[0])
        self.dec_min = float(min(dec[1:]) + dec[0])

        self.logger.info(f"Coordinates: RA = {ra[0]} ({self.ra_min} - {self.ra_max})")
        self.logger.info(
            f"Coordinates: Dec = {dec[0]} ({self.dec_min} - {self.dec_max})"
        )

        self.summary_path = f"{nu_candidate_output_dir}/{nu_name}.pdf"
        BaseScanner.__init__(
            self,
            t_min=nu_time,
            run_config=NU_RUN_CONFIG,
            logger=logger,
            cone_nside=cone_nside,
        )
        self.prob_threshold = 0.9
        self.area = (
            (self.ra_max - self.ra_min)
            * (self.dec_max - self.dec_min)
            * abs(np.cos(np.radians(dec[0])))
        )
        self.logger.info(f"Projected Area: {self.area}")
        (
            self.map_coords,
            self.pixel_nos,
            self.nside,
            self.map_probs,
            self.data,
            self.key,
        ) = self.unpack_map()

    def get_name(self):
        """ """
        return self.nu_name

    def get_full_name(self):
        """ """
        return f"neutrino event {self.get_name()} ({self.author} et. al, GCN {self.gcn_no})"

    def get_overlap_line(self):
        """ """
        return (
            f"We covered {self.area:.1f} sq deg, corresponding to {self.overlap_prob:.1f}% of the reported localization region. "
            "This estimate accounts for chip gaps. "
        )

    def candidate_text(
        self, ztf_id: str, first_detection: float, lul_lim: float, lul_jd: float
    ):
        """ """
        fd = Time(first_detection, format="mjd")

        text = f"{ztf_id} was first detected on {fd.utc}. "

        return text

    @staticmethod
    def get_obs_line():
        """ """
        return "Each exposure was 300s with a typical depth of 21.0 mag."

    @staticmethod
    def remove_variability_line():
        """ """
        return ""

    def filter_f_no_prv(self, res: dict):
        """ """

        ztf_id = res["objectId"]

        # Positive detection
        if res["candidate"]["isdiffpos"] not in ["t", "1"]:
            self.logger.debug(f"{ztf_id}: Negative subtraction")
            return False

        try:
            if res["candidate"]["drb"] < 0.3:
                self.logger.debug(f"{ztf_id}: DRB too low")
                return False
        except (KeyError, TypeError) as e:
            pass

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.debug(f"{ztf_id}: Not in contour")
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            self.logger.debug(
                f"{ztf_id}: Does not have 2 detections separated  by >15 mins"
            )
            return False

        return True

    def filter_f_history(self, res: dict):
        """Filter based on 2 detection requirement and probability contour requirement"""

        ztf_id = res["objectId"]

        n_detections = len(
            [x for x in res["prv_candidates"] if "isdiffpos" in x.keys()]
        )

        if n_detections < 1:
            self.logger.debug(f"{ztf_id}: Has insufficient detection")
            return False

        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.debug(f"{ztf_id}: Not in contour")
            return False

        return True

    def find_cone_coords(self):
        """ """
        cone_coords = []
        cone_ids = []

        scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))

        self.logger.info("Finding search pixels:")

        for i in tqdm(range(hp.nside2npix(self.cone_nside))):
            ra, dec = self.extract_ra_dec(self.cone_nside, i)
            ra_rad = np.radians(ra)
            dec_rad = np.radians(dec)
            if np.logical_and(
                ra > self.ra_min - scan_radius, ra < self.ra_max + scan_radius
            ):
                if np.logical_and(
                    dec > self.dec_min - scan_radius,
                    dec < self.dec_max + scan_radius,
                ):
                    cone_coords.append((ra_rad, dec_rad))
                    cone_ids.append(i)

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", np.float), ("dec", np.float)])
        )

        return cone_ids, cone_coords

    def in_contour(self, ra_deg, dec_deg):

        in_ra = np.logical_and(ra_deg > self.ra_min, ra_deg < self.ra_max)
        in_dec = np.logical_and(dec_deg > self.dec_min, dec_deg < self.dec_max)

        return np.logical_and(in_ra, in_dec)

    def unpack_map(self):
        """ """
        # nside = self.cone_nside
        nside = 1024
        map_coords = []
        pixel_nos = []

        center_ra = np.radians(np.mean([self.ra_max, self.ra_min]))
        center_dec = np.radians(np.mean([self.dec_max, self.dec_min]))
        rad = (
            np.radians(max(self.ra_max - self.ra_min, self.dec_max - self.dec_min))
            / 2.0
        )

        nearish_pixels = list(
            hp.query_disc(
                nside=nside,
                vec=hp.ang2vec(np.pi / 2.0 - center_dec, center_ra),
                radius=rad,
                nest=True,
            )
        )

        for i in tqdm(nearish_pixels):
            ra, dec = self.extract_ra_dec(nside, i)
            if self.in_contour(ra, dec):
                map_coords.append((ra, dec))
                pixel_nos.append(i)

        map_probs = np.ones_like(pixel_nos, dtype=np.float)
        map_probs /= np.sum(map_probs)

        key = "PROB"

        data = np.zeros(hp.nside2npix(nside), dtype=np.dtype([(key, np.float)]))
        data[np.array(pixel_nos)] = map_probs

        return map_coords, pixel_nos, nside, map_probs, data, key


if __name__ == "__main__":

    import logging

    logger = logging.getLogger("quiet_logger")
    logger.setLevel(logging.ERROR)

    nu = NeutrinoScanner(logger=logger)
    nu.scan_cones()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nuztf",
    version="2.2.0",
    author="Robert Stein, Simeon Reusch, Jannis Necker",
    author_email="robert.stein@desy.de, simeon.reusch@desy.de, jannis.necker@desy.de",
    description="Package for multi-messenger correlation searches with ZTF",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="astroparticle physics science multimessenger astronomy ZTF",
    url="https://github.com/desy-multimessenger/nuztf",
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.8.0,<3.9.0",
    install_requires=[
        "ampel-alerts == 0.7.2",
        "ampel-core == 0.7.4",
        "ampel-interface == 0.7.1",
        "ampel-photometry == 0.7.1",
        "ampel-ztf == 0.7.4",
        "astrobject == 0.8.6",
        "astropy == 4.3.1",
        "astropy_healpix == 0.6",
        "backoff == 1.11.1",
        "bs4 == 0.0.1",
        "catsHTM == 0.1.32",
        "coveralls == 3.2.0",
        "datetime == 4.3",
        "extcats == 2.4.1",
        "fastavro == 1.4.5",
        "fitsio == 1.1.5",
        "gwemopt == 0.0.73",
        "healpy == 1.15.0",
        "ipykernel == 6.4.1",
        "jupyter == 1.0.0",
        "ligo-gracedb == 2.7.6",
        "lxml==4.6.3",
        "matplotlib==3.4.3",
        "numpy==1.21.2",
        "pandas == 1.3.3",
        "psycopg2-binary == 2.9.1",
        "pydantic == 1.4",
        "pymongo == 3.12.0",
        "pysedm == 0.27.4",
        "pyvo == 1.1",
        "ratelimit == 2.2.1",
        "requests == 2.26.0",
        "scipy == 1.7.1",
        "setuptools == 58.2.0",
        "shapely == 1.7.1",
        "sklearn == 0.0",
        "slackclient == 2.9.3",
        "sqlalchemy == 1.4.25",
        "tqdm == 4.62.3",
        "wget == 3.2",
        "zerorpc == 0.6.3",
        "ztf-plan-obs == 0.33",
        "ztfquery == 1.16.0",
    ],
)



import os
import logging
from ztfquery import io

# Manage ztfquery logins from environment variables


def load_credentials(name):
    """ZTFquery wrapper for loading credentials."""
    return io._load_id_(name)


try:
    io.set_account(
        "irsa", username=os.environ["IRSA_USER"], password=os.environ["IRSA_PASSWORD"]
    )
    logging.info('Set up "irsa" credentials')

except KeyError:
    logging.info(
        'No Credentials for "irsa" found in environment' "Assuming they are set."
    )

try:
    io.set_account(
        "skyvision",
        username=os.environ["SKYVISION_USER"],
        password=os.environ["SKYVISION_PASSWORD"],
    )
    logging.info('Set up "skyvision" credentials')

except KeyError:
    logging.info(
        'No Credentials for "skyvision" found in environment' "Assuming they are set."
    )

try:
    io.set_account(
        "ampel_api_archive_token",
        username=os.environ["AMPEL_API_ARCHIVE_TOKEN_USER"],
        password=os.environ["AMPEL_API_ARCHIVE_TOKEN_PASSWORD"],
    )
    logging.info('Set up "ampel_api_archive_token" credentials')

except KeyError:
    logging.info("No Token for AMPEL API found in environment" "Assume they are set.")
